"""pytest definitions to run the unittests."""
from __future__ import annotations
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Generator

import pytest
import numpy as np
import xarray as xr


def get_blobs(blobs: int, size: int) -> np.ndarray:
    """Create an array with blobs."""
    y_coords, x_coords = np.meshgrid(np.arange(100), np.arange(100))
    x_blob = np.arange(0, 100, 100 / max(blobs**2, 2))[1:]
    data_array = np.zeros((100, 100))
    centres: list[tuple[int, int]] = []
    for i in x_blob:
        centres.append((i, i))
    for centre in centres:
        is_smaller = (x_coords - centre[1]) ** 2 + (y_coords - centre[0]) ** 2 < (
            size / 2
        ) ** 2
        data_array[is_smaller] = 1
    return data_array * 10


def create_data(variable_name: str, blobs: int, size: int) -> xr.Dataset:
    """Create a netcdf dataset."""
    coords: dict[str, np.ndarray] = {}
    coords["x"] = np.linspace(-10, -5, 100)
    coords["y"] = np.linspace(120, 125, 100)
    lat, lon = np.meshgrid(coords["y"], coords["x"])
    Lon = xr.DataArray(lon, name="Lg", coords=coords, dims=("y", "x"))
    Lat = xr.DataArray(lat, name="Lt", coords=coords, dims=("y", "x"))
    coords["time"] = np.array(
        [np.datetime64("2020-01-01T00:00"), np.datetime64("2020-01-01T00:10")]
    )
    dims = (2, 100, 100)
    data_array = np.empty(dims)
    for time in range(dims[0]):
        data_array[time] = get_blobs(blobs, size)
    dset = xr.DataArray(
        data_array,
        dims=("time", "y", "x"),
        coords=coords,
        name=variable_name,
    )
    data_array = np.zeros(dims)
    return xr.Dataset({variable_name: dset, "Lt": Lat, "Lg": Lon}).set_coords(
        list(coords.keys())
    )


@pytest.fixture(scope="session")
def save_dir() -> Generator[Path, None, None]:
    """Crate a temporary directory."""
    with TemporaryDirectory() as td:
        yield Path(td)


@pytest.fixture(scope="session")
def data_with_a_blob() -> Generator[xr.Dataset, None, None]:
    """Define a simple dataset with a blob in the middle."""
    data = create_data("precip", 1, 20)
    yield data


@pytest.fixture(scope="session")
def real_data_files() -> Generator[Path, None, None]:
    """Return the path where real data resides."""
    this_dir = Path(__file__).absolute().parent
    data_dir = this_dir.parent.parent.parent
    yield data_dir / "docs" / "source" / "_static" / "data"


@pytest.fixture(scope="session")
def netcdf_files_with_blob(
    data_with_a_blob: xr.Dataset,
) -> Generator[Path, None, None]:
    """Save data with a blob to file."""

    with TemporaryDirectory() as td:
        out_file = Path(td) / "out_f.nc"
        data_with_a_blob.to_netcdf(out_file)
        yield out_file



"""Tests for various utilities."""

from datetime import timedelta

import pandas as pd
import numpy as np
import xarray as xr


def test_time_interval() -> None:
    """Test for getting the time interval."""

    from tintx.helpers import get_interval

    dt_1 = timedelta(seconds=3600)
    time1 = np.datetime64("2020-01-01T00:00")
    time2 = np.datetime64("2020-01-01T01:00")
    dt_2 = time2 - time1
    assert get_interval(dt_1) == get_interval(dt_2) == 3600


def test_shift() -> None:
    """Test the phase corrleation function."""

    from tintx.phase_correlation import get_global_shift

    shift_array = np.ones((10, 10))
    assert get_global_shift(None, None) is None
    assert get_global_shift(None, shift_array) is None
    assert get_global_shift(shift_array, None) is None
    assert isinstance(get_global_shift(shift_array, shift_array), np.ndarray)


def test_meta_reader(data_with_a_blob: xr.Dataset) -> None:
    """Test the metdata reader."""
    from tintx.helpers import MetaData

    metdata = MetaData(
        data_with_a_blob,
        "precip",
        {"x_coord": "Lg", "y_coord": "Lt", "time_coord": "time"},
    )
    assert isinstance(metdata.to_dataframe("time"), pd.Series)
    assert isinstance(metdata.to_dataframe("Lg"), pd.DataFrame)



"""Test data vis."""

from pathlib import Path
from tempfile import TemporaryDirectory
import pytest
import xarray as xr


def test_animation(data_with_a_blob: xr.Dataset) -> None:
    """Simple test for creating animations."""

    from tintx import RunDirectory

    rd = RunDirectory(data_with_a_blob, "precip", x_coord="x", y_coord="y")
    _ = rd.get_tracks(field_thresh=0)
    animation = rd.animate(tracers=True, plot_style={"title": "test"})
    with TemporaryDirectory() as temp_dir:
        animation.save(Path(temp_dir) / "test.gif")
        assert (Path(temp_dir) / "test.gif").is_file()


def test_plot_tracks(data_with_a_blob: xr.Dataset) -> None:
    """Simple test for plotting tracks."""
    from tintx import RunDirectory

    rd = RunDirectory(data_with_a_blob, "precip", x_coord="x", y_coord="y")
    _ = rd.get_tracks(field_thresh=1000)
    with pytest.raises(TypeError):
        _ = rd.plot_trajectories(ax=1)
    with pytest.raises(ValueError):
        _ = rd.plot_trajectories()


def test_plot_real_data(real_data_files: Path) -> None:
    """Run the tracking with real data and plot it."""
    from tintx import RunDirectory
    from cartopy.mpl.geoaxes import GeoAxesSubplot

    files = list(real_data_files.rglob("*CPOL*.nc"))
    first = "2006-11-16 03:00"
    last = "2006-11-16 11:00"
    rd = RunDirectory.from_files(
        files,
        "radar_estimated_rain_rate",
        x_coord="longitude",
        y_coord="latitude",
        start=first,
        end=last,
    )
    _ = rd.get_tracks(field_thresh=0.001, min_size=0)
    ax = rd.plot_trajectories(label=True)
    assert isinstance(ax, GeoAxesSubplot)


def test_plot_kwargs() -> None:
    """Test the keyword correction of plotting."""
    from tintx.visualization import _normalize_kwargs

    kwargs = _normalize_kwargs(
        {
            "c": "yellow",
        }
    )
    assert "color" in kwargs
    assert kwargs["color"] == "yellow"

    kwargs = _normalize_kwargs({"ms": 20}, kind="line2d")
    assert "markersize" in kwargs
    assert kwargs["markersize"] == 20



"""
Tracking tuning parameters
--------------------------

The following parameter can be set to tune the cell tracking algorithm

* field_thresh: units of 'field' attribute, default: 32
    The threshold used for object detection. Detected objects are connected
    pixels above this threshold.
* iso_thresh: units of 'field' attribute, default: 4
    Used in isolated cell classification. Isolated cells must not be connected
    to any other cell by contiguous pixels above this threshold.
* iso_smooth: pixels, default: 4
    Gaussian smoothing parameter in peak detection preprocessing. See
    single_max in tint.objects.
* min_size: square kilometers, default: 8
    The minimum size threshold in pixels for an object to be detected.
* search_margin: meters, default: 250
    The radius of the search box around the predicted object center.
* flow_margin: meters, default: 750
    The margin size around the object extent on which to perform phase
    correlation.
* max_disparity: float, default: 999
    Maximum allowable disparity value. Larger disparity values are sent to a
    large number.
* max_flow_mag: meters per second, default: 50
    Maximum allowable global shift magnitude. See get_global_shift in
    tint.phase_correlation.
* max_shift_disp: meters per second, default: 15
    Maximum magnitude of difference in meters per second for two shifts to be
    considered in agreement. See correct_shift in tint.matching.
* gs_alt: meters, default: 1500
    Altitude in meters at which to perform phase correlation for global shift
    calculation. See correct_shift in tint.matching.


Setting and getting the tuning parameters
-----------------------------------------
The parameters can the set using the :py:class:`tintx.config.set` class.
Getting the values of the currently set parameters can be done by the
:py:func:`tintx.config.get` method.
"""

from __future__ import annotations
from typing import cast, Any, Optional
from typing_extensions import Literal

config: dict[str, float] = dict(
    ISO_THRESH=4.0,
    FIELD_THRESH=32.0,
    ISO_SMOOTH=4.0,
    MIN_SIZE=8.0,
    SEARCH_MARGIN=250.0,
    FLOW_MARGIN=750.0,
    MAX_DISPARITY=999.0,
    MAX_FLOW_MAG=50.0,
    MAX_SHIFT_DISP=15.0,
    GS_ALT=1500.0,
)
global_config: dict[str, float] = config  # alias


def get(key: str, default: Optional[float] = None) -> Optional[float]:
    """
    Get elements from global tinitX config


    Examples
    --------
    .. execute_code::
        :hide_headers:

        from tintx import config
        print(config.get('min_size'))

    .. execute_code::
        :hide_headers:

        from tintx import config
        print(config.get('bar', default=123.))


    See Also
    --------
    tintx.config.set
    """

    return config.get(key.upper(), default)


class set:
    """Set tint configuration values within a context manager

    Parameters
    ----------
    **kwargs : float
        key-value pairs of the config values to set.

    Example
    -------
    Set ``field_thresh`` parameter in a context. After the ``with`` block
    the parameter will be set back to its original value.

    .. execute_code::
        :hide_headers:

        from tintx import config
        with config.set(field_thresh=5):
            print(config.get("field_thresh"))
        print(config.get("field_thresh"))

    To make the configuration changes persistent you can use ``set`` without
    the ``with`` block:

    .. execute_code::
        :hide_headers:

        import tintx
        tintx.config.set(field_thresh=5)


    See Also
    --------
    tintx.config.get
    """

    config: dict[str, float]
    _record: list[tuple[Literal["insert", "replace"], str, Optional[float]]]

    def __init__(
        self,
        **kwargs: float,
    ):
        self.config = config
        self._record = []

        for key, value in kwargs.items():
            self._assign(key.upper(), float(value), config)

    def __enter__(self) -> set:
        return self

    def __exit__(self, *args: Any) -> None:
        for op, key, value in reversed(self._record):
            if op == "replace":
                self.config[key] = cast(float, value)
            else:  # insert
                self.config.pop(key, None)

    def get(self, key: str, default: Optional[float] = None) -> Optional[float]:
        """
        Convenience method to retrieve elements from the ``tinitx`` configuration


        Examples
        --------
        .. execute_code::
            :hide_headers:

            from tintx import config
            with config.set(field_thresh=5) as cfg:
                print(cfg.get("field_thresh"))

        """
        return get(key, default)

    def _assign(
        self,
        key: str,
        value: float,
        cfg: dict[str, float],
        record: bool = True,
    ) -> None:
        """Assign value into a nested configuration dictionary

        Parameters
        ----------
        keys : Sequence[str]
            The nested path of keys to assign the value.
        value : object
        d : dict
            The part of the nested dictionary into which we want to assign the
            value
        path : tuple[str], optional
            The path history up to this point.
        record : bool, optional
            Whether this operation needs to be recorded to allow for rollback.
        """
        if record:
            if key in cfg:
                self._record.append(("replace", key, cfg[key]))
            else:
                self._record.append(("insert", key, None))
        cfg[key] = value



"""
tint.phase_correlation
======================

Functions for performing phase correlation. Used to predict cell movement
between scans.

"""
from __future__ import annotations
from typing import Optional, overload, Union
from typing_extensions import Literal

import numpy as np
from scipy import ndimage

from .types import ConfigType


def get_ambient_flow(
    obj_extent: dict[str, np.ndarray],
    img1: np.ndarray,
    img2: np.ndarray,
    params: ConfigType,
    grid_size: Union[np.ndarray, tuple[int, int, int]],
) -> Optional[float]:
    """Takes in object extent and two images and returns ambient flow. Margin
    is the additional region around the object used to compute the flow
    vectors."""
    margin_r = params["FLOW_MARGIN"] / grid_size[1]
    margin_c = params["FLOW_MARGIN"] / grid_size[2]
    row_lb = obj_extent["obj_center"][0] - obj_extent["obj_radius"] - margin_r
    row_ub = obj_extent["obj_center"][0] + obj_extent["obj_radius"] + margin_r
    col_lb = obj_extent["obj_center"][1] - obj_extent["obj_radius"] - margin_c
    col_ub = obj_extent["obj_center"][1] + obj_extent["obj_radius"] + margin_c
    row_lb = int(row_lb)
    row_ub = int(row_ub)
    col_lb = int(col_lb)
    col_ub = int(col_ub)

    dims = img1.shape

    row_lb = np.max([row_lb, 0])
    row_ub = np.min([row_ub, dims[0]])
    col_lb = np.max([col_lb, 0])
    col_ub = np.max([col_ub, dims[1]])

    flow_region1 = np.copy(img1[row_lb : row_ub + 1, col_lb : col_ub + 1])
    flow_region2 = np.copy(img2[row_lb : row_ub + 1, col_lb : col_ub + 1])

    flow_region1[flow_region1 != 0] = 1
    flow_region2[flow_region2 != 0] = 1
    return fft_flowvectors(flow_region1, flow_region2)


def fft_flowvectors(
    im1: np.ndarray, im2: np.ndarray, global_shift: bool = False
) -> Optional[float]:
    """Estimates flow vectors in two images using cross covariance."""
    if not global_shift and (np.max(im1) == 0 or np.max(im2) == 0):
        return None

    crosscov = fft_crosscov(im1, im2)
    sigma = (1 / 8) * min(crosscov.shape)
    cov_smooth = ndimage.gaussian_filter(crosscov, sigma)
    dims = np.array(im1.shape)
    pshift = np.argwhere(cov_smooth == np.max(cov_smooth))[0]
    pshift = (pshift + 1) - np.round(dims / 2, 0)
    return pshift


def fft_crosscov(im1: np.ndarray, im2: np.ndarray) -> np.ndarray:
    """Computes cross correlation matrix using FFT method."""
    fft1_conj = np.conj(np.fft.fft2(im1))
    fft2 = np.fft.fft2(im2)
    normalize = abs(fft2 * fft1_conj)
    normalize[normalize == 0] = 1  # prevent divide by zero error
    cross_power_spectrum = (fft2 * fft1_conj) / normalize
    crosscov = np.fft.ifft2(cross_power_spectrum)
    crosscov = np.real(crosscov)
    return fft_shift(crosscov)


def fft_shift(fft_mat: np.ndarray) -> np.ndarray:
    """Rearranges the cross correlation matrix so that 'zero' frequency or DC
    component is in the middle of the matrix. Taken from stackoverflow Que.
    30630632."""
    rd2 = int(fft_mat.shape[0] / 2)
    cd2 = int(fft_mat.shape[1] / 2)
    quad1 = fft_mat[:rd2, :cd2]
    quad2 = fft_mat[:rd2, cd2:]
    quad3 = fft_mat[rd2:, cd2:]
    quad4 = fft_mat[rd2:, :cd2]
    centered_t = np.concatenate((quad4, quad1), axis=0)
    centered_b = np.concatenate((quad3, quad2), axis=0)
    centered = np.concatenate((centered_b, centered_t), axis=1)
    return centered


@overload
def get_global_shift(im1: Literal[None], im2: Literal[None]) -> None:
    ...  # pragma: no cover


@overload
def get_global_shift(im1: np.ndarray, im2: Literal[None]) -> None:
    ...  # pragma: no cover


@overload
def get_global_shift(im1: Literal[None], im2: np.ndarray) -> None:
    ...  # pragma: no cover


@overload
def get_global_shift(im1: np.ndarray, im2: np.ndarray) -> float:
    ...  # pragma: no cover


def get_global_shift(
    im1: Optional[np.ndarray], im2: Optional[np.ndarray]
) -> Optional[float]:
    """Returns standardazied global shift vector. im1 and im2 are full frames
    of raw DBZ values."""
    if im2 is None or im1 is None:
        return None
    shift = fft_flowvectors(im1, im2, global_shift=True)
    return shift



"""Test the highlevel tracking results."""

from pathlib import Path
from tempfile import NamedTemporaryFile

import numpy as np
import pandas as pd
import pytest
import xarray as xr


def test_tracking_of_dataset(data_with_a_blob: xr.Dataset) -> None:
    """Test the tracking of fake data with one field of rainfall."""

    from tintx import RunDirectory

    run_dir = RunDirectory(data_with_a_blob, "precip", x_coord="Lg", y_coord="Lt")
    run_dir2 = RunDirectory(
        data_with_a_blob["precip"], "precip", x_coord="x", y_coord="x"
    )
    assert isinstance(run_dir2.data, xr.Dataset)
    assert len(run_dir.tracks) == 0
    tracks = run_dir.get_tracks(field_thresh=0.0)
    assert tracks == 1
    assert len(run_dir.tracks) == 2
    tracks = run_dir.get_tracks(field_thresh=1000)
    assert tracks == 0


def test_get_parameters(data_with_a_blob: xr.DataArray) -> None:
    """Test getting the tuning parameters."""

    from tintx import RunDirectory

    run_dir = RunDirectory(data_with_a_blob, "precip", x_coord="Lg", y_coord="Lt")
    _ = run_dir.get_tracks(field_thresh=0.0)
    assert run_dir.get_parameters(run_dir.tracks) == run_dir.get_parameters()
    with pytest.raises(ValueError):
        run_dir.get_parameters(pd.DataFrame({"foo": ["bar"]}))


def test_tracking_of_datafile(netcdf_files_with_blob: Path) -> None:
    """Test tracking from saved netcdf files."""

    from tintx import RunDirectory

    run_dir = RunDirectory.from_files(
        netcdf_files_with_blob, "precip", x_coord="x", y_coord="y"
    )
    assert len(run_dir.tracks) == 0
    tracks = run_dir.get_tracks(field_thresh=0.0, iso_thresh=0)
    assert tracks == 1
    _ = run_dir.get_tracks(field_thresh=1000, flush=False)
    assert len(run_dir.tracks) == tracks + 1


def test_save_dataset(save_dir: Path, netcdf_files_with_blob: Path) -> None:
    """Test saving the dataset."""

    from tintx import RunDirectory

    run_dir = RunDirectory.from_files(
        netcdf_files_with_blob, "precip", x_coord="x", y_coord="y"
    )
    _ = run_dir.get_tracks(field_thresh=0.0, iso_thresh=0)
    run_dir.save_tracks(save_dir / "test.h5")
    assert len(run_dir.tracks) == len(
        pd.read_hdf(save_dir / "test.h5", "tintx_tracks")
    )


def test_load_dataset(save_dir: Path, netcdf_files_with_blob: Path) -> None:
    """Test loading the data from dataframe."""

    from tintx import RunDirectory

    save_file = save_dir / "test.h5"
    run_dir2 = RunDirectory.from_files(
        netcdf_files_with_blob, "precip", x_coord="x", y_coord="y"
    )
    _ = run_dir2.get_tracks(field_thresh=0.0)
    run_dir2.save_tracks(save_file)
    run_dir1 = RunDirectory.from_dataframe(save_file)
    run_dir3 = RunDirectory.from_dataframe(save_file, run_dir2.data)
    run_dir3._files = ""
    run_dir3.save_tracks(save_file)
    assert len(run_dir1.tracks) == len(run_dir2.tracks) == len(run_dir3.tracks)


def test_load_empty(data_with_a_blob: xr.DataArray) -> None:
    """Test loading empty datasets."""
    from tintx import RunDirectory

    run_dir = RunDirectory(data_with_a_blob, "precip", x_coord="Lg", y_coord="Lt")
    _ = run_dir.get_tracks(field_thresh=0.0)
    with NamedTemporaryFile(suffix=".h5") as save_file:
        run_dir.save_tracks(save_file.name)
        with pytest.warns(UserWarning):
            run_dir2 = RunDirectory.from_dataframe(save_file.name)
    data1 = run_dir.data
    data2 = run_dir2.data
    for dim in data1.dims:
        assert dim in data2.dims
        if dim != "time":
            assert np.allclose(data1[dim].values, data2[dim].values)
    assert "Lg" in data2.coords
    assert "Lt" in data2.coords
    assert "precip" in data2.data_vars
    assert data2["precip"].shape == data1["precip"].shape



"""Test for the command line interface."""

from pathlib import Path
from tempfile import TemporaryDirectory

import pytest


def test_help() -> None:
    """Test the help."""

    from tintx.cli import tintx

    with pytest.raises(SystemExit):
        tintx()


def test_tracking_and_plotting(netcdf_files_with_blob: Path) -> None:
    """Test plotting and tracking."""

    from tintx.cli import tintx

    with TemporaryDirectory() as td:
        save_file = Path(td) / "test.h5"
        tintx(
            [
                "track",
                "precip",
                str(netcdf_files_with_blob),
                "--x-coord=x",
                "--y-coord=y",
                "--field-thresh=0",
                f"--output={save_file}",
            ],
            standalone_mode=False,
        )
        assert save_file.is_file()
        tintx(
            [
                "plot",
                str(save_file),
                "--animate",
                f"-o={save_file.with_suffix('.mp4')}",
            ],
            standalone_mode=False,
        )
        tintx(
            [
                "plot",
                str(save_file),
                f"-o={save_file.with_suffix('.png')}",
            ],
            standalone_mode=False,
        )


def test_get_files(netcdf_files_with_blob: Path) -> None:
    """Test getting the files."""

    from tintx.cli import _get_file_names

    wild_card = (netcdf_files_with_blob.parent / "*.nc",)
    files1 = tuple(_get_file_names((netcdf_files_with_blob,)))
    files2 = tuple(_get_file_names((netcdf_files_with_blob.parent,)))
    files3 = tuple(_get_file_names(wild_card))
    assert files1 == files2 == files3



"""
tint.tracks
===========

Cell_tracks class.

"""

from __future__ import annotations
import copy
from typing import cast, Dict, Iterator, Optional

import pandas as pd
import numpy as np
from tqdm.std import tqdm

from .grid_utils import get_grid_size, get_radar_info, extract_grid_data
from .helpers import Record, Counter, GridType
from .phase_correlation import get_global_shift
from .matching import get_pairs
from .objects import init_current_objects, update_current_objects
from .objects import get_object_prop, write_tracks
from .config import config as tint_config
from .types import ConfigType


class Cell_tracks:
    """
    This is the main class in the module. It allows tracks
    objects to be built using lists of data arrays.

    Attributes
    ----------
    params : dict
        Parameters for the tracking algorithm.
    field : str
        String specifying data variable to be used for tracking. Default is
        'reflectivity'.
    grid_size : array
        Array containing z, y, and x mesh size in meters respectively.
    last_grid : Grid
        Contains the most recent grid object tracked. This is used for dynamic
        updates.
    counter : Counter
        See Counter class.
    record : Record
        See Record class.
    current_objects : dict
        Contains information about objects in the current scan.
    _tracks : DataFrame

    _saved_record : Record
        Deep copy of Record at the penultimate scan in the sequence. This and
        following 2 attributes used for link-up in dynamic updates.
    _saved_counter : Counter
        Deep copy of Counter.
    _saved_objects : dict
        Deep copy of current_objects.

    """

    def __init__(self, field: str = "reflectivity"):

        self.field = field
        self.grid_size: Optional[np.ndarray] = None
        self.radar_info: Optional[dict[str, float]] = None
        self.last_grid: Optional[GridType] = None
        self.counter: Optional[Counter] = None
        self.record: Optional[Record] = None
        self.current_objects: Optional[dict[str, np.ndarray]] = None
        self._tracks = pd.DataFrame()

        self._saved_record: Optional[Record] = None
        self._saved_counter: Optional[Counter] = None
        self._saved_objects: Optional[dict[str, np.ndarray]] = None

    @property
    def params(self) -> ConfigType:
        """Get the tracking parameters."""
        return cast(ConfigType, tint_config)

    def _save(self) -> None:
        """Saves deep copies of record, counter, and current_objects."""
        self._saved_record = copy.deepcopy(self.record)
        self._saved_counter = copy.deepcopy(self.counter)
        self._saved_objects = copy.deepcopy(self.current_objects)

    def _load(self) -> None:
        """Loads saved copies of record, counter, and current_objects. If new
        tracks are appended to existing tracks via the get_tracks method, the
        most recent scan prior to the addition must be overwritten to link up
        with the new scans. Because of this, record, counter and
        current_objects must be reverted to their state in the penultimate
        iteration of the loop in get_tracks. See get_tracks for details."""
        self.record = self._saved_record
        self.counter = self._saved_counter
        self.current_objects = self._saved_objects

    @property
    def tracks(self) -> pd.DataFrame:
        """A pandas.DataFrame representation of the tracked cells."""
        return self._tracks

    def _get_tracks(
        self,
        grids: Iterator[GridType],
        pbar: tqdm,
        centre: Optional[tuple[float, float]] = None,
    ) -> int:
        raw2: Optional[np.ndarray] = None
        if self.record is None:
            # tracks object being initialized
            grid_obj2 = next(grids)
            self.grid_size = get_grid_size(grid_obj2)
            if centre is None:
                xgrid = grid_obj2.x.values
                ygrid = grid_obj2.y.values
                if len(xgrid.shape) == 2:
                    x_c = xgrid[xgrid.shape[0] // 2][xgrid.shape[1] // 2]
                else:
                    x_c = xgrid[xgrid.shape[0] // 2]
                if len(ygrid.shape) == 2:
                    y_c = ygrid[ygrid.shape[0] // 2][ygrid.shape[1] // 2]
                else:
                    y_c = ygrid[ygrid.shape[0] // 2]
                x_c = cast(float, x_c)
                y_c = cast(float, y_c)
                self.radar_info = get_radar_info((x_c, y_c))
            else:
                self.radar_info = get_radar_info(centre)
            self.counter = Counter()
            self.record = Record(grid_obj2)
        else:
            # tracks object being updated
            grid_obj2 = cast(GridType, self.last_grid)
            self._tracks.drop(self.record.scan + 1)  # last scan is overwritten

        new_rain = bool(self.current_objects is None)
        stop_iteration = bool(grid_obj2 is None)
        raw2, frame2 = extract_grid_data(grid_obj2, self.params)
        while not stop_iteration:
            pbar.update()
            grid_obj1 = grid_obj2
            raw1 = raw2
            frame1 = frame2

            try:
                grid_obj2 = next(grids)
            except StopIteration:
                stop_iteration = True

            if not stop_iteration:
                self.record.update_scan_and_time(grid_obj1, grid_obj2)
                raw2, frame2 = extract_grid_data(grid_obj2, self.params)
            else:
                # setup to write final scan
                self._save()
                self.last_grid = grid_obj1
                self.record.update_scan_and_time(grid_obj1)
                raw2 = None
                frame2 = np.zeros_like(frame1)

            if np.nanmax(frame1) == 0:
                new_rain = True
                self.current_objects = None
                continue
            global_shift = cast(float, get_global_shift(raw1, raw2))
            pairs = cast(
                np.ndarray,
                get_pairs(
                    frame1,
                    frame2,
                    global_shift,
                    self.current_objects,
                    self.record,
                    self.params,
                ),
            )
            if new_rain:
                # first nonempty scan after a period of empty scans
                self.current_objects, self.counter = init_current_objects(
                    frame1, frame2, pairs, cast(Counter, self.counter)
                )
                new_rain = False
            else:
                self.current_objects, self.counter = update_current_objects(
                    frame1,
                    frame2,
                    pairs,
                    cast(Dict[str, np.ndarray], self.current_objects),
                    cast(Counter, self.counter),
                )
            obj_props = get_object_prop(
                frame1, grid_obj1, self.record, self.params
            )

            self.record.add_uids(self.current_objects)
            self._tracks = write_tracks(
                self._tracks, self.record, self.current_objects, obj_props
            )
            del grid_obj1, raw1, frame1, global_shift, pairs, obj_props
            # scan loop end
        self._load()
        ncells = 0
        if len(self._tracks):
            ncells = self._tracks.index.get_level_values(1).astype(int).max() + 1
        return ncells



"""Tests for the tint configuration."""

import xarray as xr


def test_set_config(data_with_a_blob: xr.Dataset) -> None:
    """Test the set class."""

    from tintx import config, RunDirectory

    with config.set(field_thresh=0.0, foo=2) as cfg:
        run_dir = RunDirectory(
            data_with_a_blob, "precip", x_coord="Lg", y_coord="Lt"
        )

        assert run_dir.params["FIELD_THRESH"] == 0
        assert cfg.get("foo") == 2
    assert config.get("foo", 3) == 3
    assert run_dir.params["FIELD_THRESH"] != 0


def test_get_config() -> None:

    from tintx import config

    assert config.get("FIELD_THRESH") == config.get("field_thresh") == 32.0
    assert config.get("foo") is None
    assert config.get("foo", default=23) == 23



"""Command line interface for tintX tracking."""
from __future__ import annotations
from pathlib import Path
from typing import Iterator, Optional

import click
from tintx import __version__, RunDirectory, config


CONTEXT_SETTINGS = dict(
    help_option_names=["-h", "--help"],
)
PROG_NAME = "tintx"


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(__version__, "-V", "--version", prog_name=PROG_NAME)
def tintx(argv: Optional[list[str]] = None, standalone_mode: bool = True) -> None:
    """Command line interface (cli) of the tintX tracking algorithm.

    The cli offers two sub commands. One for applying the tracking algorithm,
    one for visualisation of already tracked data.
    """
    return


def _get_file_names(input_paths: tuple[Path, ...]) -> Iterator[Path]:

    suffixes = (".nc", ".nc4", ".grb", ".grib")
    for input_path in input_paths:
        if input_path.is_file() and input_path.suffix in suffixes:
            yield input_path
        elif input_path.is_dir():
            for input_file in input_path.rglob("*.*"):
                if input_file.suffix in suffixes:
                    yield input_file
        else:
            # This is a long shot, we assume that a glob pattern was given
            for input_file in input_path.parent.rglob(input_path.name):
                yield input_file


@tintx.command()
@click.argument(
    "input_file",
    nargs=1,
    type=click.Path(
        resolve_path=True, path_type=Path, exists=True, dir_okay=False
    ),
)
@click.option(
    "-o",
    "--output",
    type=click.Path(resolve_path=True, path_type=Path, dir_okay=False),
    default=None,
    help=(
        "Path to a visualisation of the tracking. If `None` is given, no "
        "visualisation will be created. The type of visualisation will be "
        "determined from the file type. E.i for `.png` or `.jpg` files "
        "trajectory plots will be created for `.mp4`, `.gif` animations "
        "will be created."
    ),
)
@click.option("--animate", is_flag=True, help="Create animation")
@click.option(
    "--dt",
    type=click.FLOAT,
    default=0,
    help="Offset in hours from UTC. Animation only",
)
@click.option(
    "--fps",
    type=click.FLOAT,
    default=2,
    help="Play back speed of an animation. Animation only",
)
@click.option(
    "--cmap",
    type=click.STRING,
    default="Blues",
    help="Colormap used for animations.",
)
@click.option(
    "--vmin", type=click.FLOAT, default=0, help="minimum values to be plotted"
)
@click.option(
    "--vmax",
    type=click.FLOAT,
    default=3,
    help="Maximum values to display. Animation only",
)
@click.option(
    "--mintrace",
    type=click.FLOAT,
    default=2,
    help="Minimum length of a trace to be plotted: Trajectory plot only",
)
@click.option(
    "--markersize",
    "-ms",
    type=click.FLOAT,
    default=25,
    help="Help marker size of the trijectory plot",
)
@click.option(
    "--linewidth",
    "-lw",
    type=click.FLOAT,
    default=1,
    help="Line width to be plotted.",
)
def plot(
    input_file: Path,
    output: Optional[Path] = None,
    cmap: str = "Blues",
    animate: bool = False,
    **kwargs: float,
) -> None:
    """Plot/Animate existing tracking data.

    Arguments:

    input_files:
        Filename of the HDF5 file contaning the tracking data.

    """

    run_dir = RunDirectory.from_dataframe(input_file)
    time_suffix = (
        f'{run_dir.start.strftime("%Y%m%dT%H%M")}-'
        f'{run_dir.end.strftime("%Y%m%dT%H%M")}'
    )
    outf = Path(".").absolute() / f"tintx_tracks_{run_dir.var_name}_{time_suffix}"
    if animate:
        output = output or outf.with_suffix(".mp4")
        output.parent.mkdir(exist_ok=True, parents=True)
        anim = run_dir.animate(
            cmap=cmap,
            vmin=kwargs["vmin"],
            vmax=kwargs["vmax"],
            dt=kwargs["dt"],
            fps=kwargs["fps"],
            plot_style={"lw": kwargs["linewidth"]},
        )
        anim.save(output, fps=kwargs["fps"])

    else:
        output = output or outf.with_suffix(".png")
        output.parent.mkdir(exist_ok=True, parents=True)
        axes = run_dir.plot_trajectories(
            thresh=kwargs["vmin"],
            mintrace=int(kwargs["mintrace"]),
            plot_style={"lw": kwargs["linewidth"], "ms": kwargs["markersize"]},
        )
        axes.figure.savefig(output, bbox_inches="tight")


@tintx.command()
@click.argument("variable", nargs=1, type=click.STRING)
@click.argument(
    "input_files",
    nargs=-1,
    type=click.Path(resolve_path=True, path_type=Path),
)
@click.option(
    "--start",
    "-s",
    type=click.STRING,
    default=None,
    help="ISO-8601 string representation of the first tracking time step.",
)
@click.option(
    "--end",
    "-e",
    type=click.STRING,
    default=None,
    help="ISO-8601 string representation of the last tracking time step.",
)
@click.option(
    "--x-coord",
    type=click.STRING,
    default="lon",
    help="Name of the X (eastward) coordinate",
)
@click.option(
    "--y-coord",
    type=click.STRING,
    default="lat",
    help="Name of the Y (northward) coordinate",
)
@click.option(
    "--time-coord",
    type=click.STRING,
    default="time",
    help="Name of the time coordinate",
)
@click.option(
    "-o",
    "--output",
    type=click.Path(resolve_path=True, path_type=Path),
    default=None,
    help=(
        "Output file where the tracking results pandas DataFrame is saved "
        "to. If `None` given (default) the output filename will be set from "
        "the input meta data and the variable name."
    ),
)
@click.option(
    "--field-thresh",
    type=click.FLOAT,
    default=32,
    help=(
        "Threshold used for object detection. Detected objects are "
        "connected pixels above this threshold."
    ),
)
@click.option(
    "--iso-thresh",
    type=click.FLOAT,
    default=4.0,
    help=(
        "Used in isolated cell classification. Isolated cells must not be  "
        "connected to any other cell by contiguous pixels above this "
        "threshold."
    ),
)
@click.option(
    "--min-size",
    type=click.FLOAT,
    default=8.0,
    help=("Minimum size threshold in pixels for an object to be detected."),
)
@click.option(
    "--search-margin",
    type=click.FLOAT,
    default=250.0,
    help=("Radius of the search box around the predicted object center."),
)
@click.option(
    "--flow-margin",
    type=click.FLOAT,
    default=750,
    help=("Margin size around objects to perform phase correlation."),
)
@click.option(
    "--max-disparity",
    type=click.FLOAT,
    default=999.0,
    help=("Maximum allowable disparity value."),
)
@click.option(
    "--max-flow-mag",
    type=click.FLOAT,
    default=50.0,
    help=("Maximum allowable global shift magnitude."),
)
@click.option(
    "--max-shift-disp",
    type=click.FLOAT,
    default=15.0,
    help=(
        "Maximum magnitude of difference in meters per second for two shifts"
        " to be considered in agreement."
    ),
)
@click.option(
    "--gs-alt",
    type=click.FLOAT,
    default=1500.0,
    help=(
        "Altitude in meters at which to perform phase correlation for global"
        "shift calculation. 3D data only"
    ),
)
@click.option(
    "--iso-smooth",
    type=click.FLOAT,
    default=4.0,
    help=("Gaussian smoothing parameter in peak detection preprocessing."),
)
def track(
    variable: str,
    input_files: tuple[Path, ...],
    start: Optional[str] = None,
    end: Optional[str] = None,
    x_coord: str = "lon",
    y_coord: str = "lat",
    time_coord: str = "time",
    output: Optional[Path] = None,
    **parameters: float,
) -> None:
    """Apply the tintX tracking algorithm.

    The sub command takes at least two arguments and attempts of read data
    saved in netcdf/grib format and apply the tracking to a data variable
    within the dataset.

    Arguments:

    variable:
        Variable name of the data that is tracked

    input_files:
        Filename(s) or Directory where the data is stored.
    """
    with config.set(**parameters):
        run_d = RunDirectory.from_files(
            _get_file_names(input_files),
            variable,
            start=start,
            end=end,
            x_coord=x_coord,
            y_coord=y_coord,
            time_coord=time_coord,
            use_cftime=True,
            combine="by_coords",
        )
        time_suffix = (
            f'{run_d.start.strftime("%Y%m%dT%H%M")}-'
            f'{run_d.end.strftime("%Y%m%dT%H%M")}'
        )
        outf = (
            Path(".").absolute() / f"tintx_tracks_{variable}_{time_suffix}.hdf5"
        )
        output = output or outf
        num_tracks = run_d.get_tracks()
        click.echo(f"Found and tracked {num_tracks} objects.")
        output.parent.mkdir(exist_ok=True, parents=True)
        run_d.save_tracks(output)



"""
tint.helpers
============

"""
from __future__ import annotations
from datetime import datetime, timedelta
from typing import Iterator, Optional, Union, cast

import cftime
from dask import array as dask_array
import numpy as np
import pandas as pd
import xarray as xr
from .grid_utils import parse_grid_datetime, get_grid_size, GridType


class MetaData:
    """Class that writes meta data to an HDF5 file."""

    time_units: str = "seconds since 1970-01-01T00:00:00"

    def __init__(
        self,
        dataset: xr.Dataset,
        variable: str,
        coords: dict[str, str],
    ) -> None:

        self._time = coords["time_coord"]
        self._x_coord = coords["x_coord"]
        self._y_coord = coords["y_coord"]

        self.dataset = dataset
        self.x_coord = dataset[coords["x_coord"]].values
        self.y_coord = dataset[coords["y_coord"]].values
        self.variable = variable

    @property
    def dims(self) -> tuple[str, ...]:
        """Get the dimensions of the DataArray."""
        return tuple(map(str, self.dataset[self.variable].dims))

    @property
    def coords(self) -> pd.DataFrame:
        """Get the coordinates of the DataArray in a pandas DataFrame."""
        coords: dict[str, pd.Series] = {
            str(v): pd.Series(self.dataset[v].values)
            for v in self.dataset[self.variable].dims
            if v != self._time
        }
        coords[self._time] = pd.Series(self.time)
        return pd.DataFrame(coords)

    @property
    def time(self) -> np.ndarray:
        """Get a number representation of the time vector in."""
        return np.array(
            [self._to_num(t) for t in self.dataset[self._time].values]
        )

    def _to_num(self, time: Union[cftime.datetime, np.datetime64]) -> int:
        cf_time = convert_to_cftime(time)
        return cftime.date2num(
            cf_time, units=self.time_units, calendar=cf_time.calendar
        )

    @property
    def calendar(self) -> str:
        cf_time = convert_to_cftime(self.dataset[self._time].values[0])
        return cf_time.calendar

    def to_dataframe(self, var_name: str) -> Union[pd.DataFrame, pd.Series]:
        """Convert the values of a dataset variable to a dataframe."""
        if len(self.dataset[var_name].shape) == 1:
            return pd.Series(self.dataset[var_name].values)
        return pd.DataFrame(self.dataset[var_name].values)

    def save(self, buffer: pd.io.pytables.Table) -> None:
        """Save attrs to a hdf5 store object."""
        self.coords.to_hdf(buffer, "dims")
        self.to_dataframe(self._x_coord).to_hdf(buffer, "x_coord")
        self.to_dataframe(self._y_coord).to_hdf(buffer, "y_coord")
        x_attrs = self.dataset[self._x_coord].attrs
        y_attrs = self.dataset[self._y_coord].attrs
        x_attrs["short_name"] = self._x_coord
        y_attrs["short_name"] = self._y_coord
        y_attrs["dims"] = self.dataset[self._y_coord].dims
        x_attrs["dims"] = self.dataset[self._x_coord].dims
        buffer.get_storer("x_coord").attrs.attrs = x_attrs
        buffer.get_storer("y_coord").attrs.attrs = y_attrs
        buffer.get_storer("x_coord").attrs.attrs = x_attrs
        buffer.get_storer("y_coord").attrs.attrs = y_attrs
        dim_attrs = {v: self.dataset[v].attrs for v in self.dims}
        dim_attrs[self._time].setdefault("calendar", self.calendar)
        dim_attrs[self._time]["units"] = self.time_units
        var_attrs = self.dataset[self.variable].attrs
        var_attrs["time_coord"] = self._time
        var_attrs["short_name"] = self.variable
        buffer.get_storer("dims").attrs.dim_attrs = dim_attrs
        buffer.get_storer("dims").attrs.var_attrs = var_attrs
        buffer.get_storer("dims").attrs.dims = self.dims

    @staticmethod
    def dataset_from_coords(buffer: pd.io.pytables.Table) -> xr.Dataset:
        """Create a xarray Dataset from metadata save in a pytable."""

        dims_df = buffer.get("dims")
        storer_obj = buffer.get_storer("dims")
        x_coords_attrs = buffer.get_storer("x_coord").attrs.attrs.copy()
        y_coords_attrs = buffer.get_storer("y_coord").attrs.attrs.copy()
        var_attrs = storer_obj.attrs.var_attrs.copy()
        var_attrs[
            "history"
        ] = f"{datetime.now().isoformat()}: Created empty dataset"
        coord_attrs = storer_obj.attrs.dim_attrs.copy()
        time_coord = var_attrs.pop("time_coord")
        metadata = buffer.get_storer("tintx_tracks").attrs.track_meta.copy()
        shape = []
        data = {}
        coords = [x_coords_attrs["short_name"], y_coords_attrs["short_name"]]
        for dim in storer_obj.attrs.dims:
            data[dim] = xr.DataArray(
                dims_df[dim].dropna().values,
                name=dim,
                dims=(dim,),
                attrs=coord_attrs[dim],
            )
            shape.append(data[dim].shape[0])
        data[time_coord].data = cftime.num2date(
            data[time_coord].data,
            coord_attrs[time_coord]["units"],
            coord_attrs[time_coord]["calendar"],
        )
        data[metadata["var_name"]] = xr.DataArray(
            dask_array.zeros(shape),
            name=metadata["var_name"],
            attrs=var_attrs,
            dims=storer_obj.attrs.dims,
        ).chunk({time_coord: 1})
        data[coords[0]] = xr.DataArray(
            buffer.get("x_coord").values,
            dims=x_coords_attrs.pop("dims"),
            name=coords[0],
            attrs=x_coords_attrs,
        )
        data[coords[1]] = xr.DataArray(
            buffer.get("y_coord").values,
            dims=y_coords_attrs.pop("dims"),
            name=coords[1],
            attrs=y_coords_attrs,
        )
        return xr.Dataset(data).set_coords(coords)


class Counter:
    """
    Counter objects generate and keep track of unique cell ids.
    Currently only the uid attribute is used, but this framework can
    accomodate further development of merge/split detection.

    Attributes
    ----------
    uid : int
        Last uid assigned.
    cid : dict
        Record of cell genealogy.

    """

    def __init__(self) -> None:
        """uid is an integer that tracks the number of independently formed
        cells. The cid dictionary keeps track of 'children' --i.e., cells that
        have split off from another cell."""
        self.uid: int = -1
        self.cid: dict[str, int] = {}

    def next_uid(self, count: int = 1) -> np.ndarray:
        """Incremented for every new independently formed cell."""
        new_uids = self.uid + np.arange(count) + 1
        self.uid += count
        return np.array([str(uid) for uid in new_uids])


class Record:
    """
    Record objects keep track of information related to the shift correction
    process.

    Attributes
    ----------
    scan : int
        Index of the current scan.
    time : datetime
        Time corresponding to scan.
    interval : timedelta
        Temporal difference between the next scan and the current scan.
    interval_ratio : float
        Ratio of current interval to previous interval.
    grid_size : array of floats
        Length 3 array containing z, y, and x mesh size in meters.
    shifts : dataframe
        Records inputs of shift correction process. See matching.correct_shift.
    new_shfits : dataframe
        Row of new shifts to be added to shifts dataframe.
    correction_tally : dict
        Tallies correction cases for performance analysis.


    Shift Correction Case Guide:
    case0 - new object, local_shift and global_shift disagree, returns global
    case1 - new object, returns local_shift
    case2 - local disagrees with last head and global, returns last head
    case3 - local disagrees with last head, returns local
    case4 - local and last head agree, returns average of both
    case5 - flow regions empty or at edge of frame, returns global_shift

    """

    def __init__(self, grid_obj: GridType) -> None:
        self.scan = -1
        self.time: Optional[Union[np.datetime64, cftime.datetime]] = None
        self.interval: Optional[Union[timedelta, np.timedelta64]] = None
        self.interval_ratio: Optional[float] = None
        self.grid_size = get_grid_size(grid_obj)
        self.shifts = pd.DataFrame()
        self.new_shifts = pd.DataFrame()
        self.correction_tally = {
            "case0": 0,
            "case1": 0,
            "case2": 0,
            "case3": 0,
            "case4": 0,
            "case5": 0,
        }

    def count_case(self, case_num: int) -> None:
        """Updates correction_tally dictionary. This is used to monitor the
        shift correction process."""
        self.correction_tally["case" + str(case_num)] += 1

    def record_shift(
        self,
        corr: float,
        gl_shift: float,
        l_heads: Optional[float],
        local_shift: float,
        case: int,
    ) -> None:
        """Records corrected shift, phase shift, global shift, and last
        heads per object per timestep. This information can be used to
        monitor and refine the shift correction algorithm in the
        correct_shift function."""
        if l_heads is None:
            l_heads = np.ma.array([-999, -999], mask=[True, True])

        new_shift_record = pd.DataFrame()
        new_shift_record["scan"] = [self.scan]
        new_shift_record["uid"] = ["uid"]
        new_shift_record["corrected"] = [corr]
        new_shift_record["global"] = [gl_shift]
        new_shift_record["last_heads"] = [l_heads]
        new_shift_record["phase"] = [local_shift]
        new_shift_record["case"] = [case]

        self.new_shifts = pd.concat([self.new_shifts, new_shift_record])

    def add_uids(self, current_objects: pd.DataFrame) -> None:
        """Because of the chronology of the get_tracks process, object uids
        cannot be added to the shift record at the time of correction, so they
        must be added later in the process."""
        if len(self.new_shifts) > 0:
            self.new_shifts["uid"] = current_objects["uid"]
            self.new_shifts.set_index(["scan", "uid"], inplace=True)
            self.shifts = pd.concat([self.shifts, self.new_shifts])
            self.new_shifts = pd.DataFrame()

    def update_scan_and_time(
        self, grid_obj1: GridType, grid_obj2: Optional[GridType] = None
    ) -> None:
        """Updates the scan number and associated time. This information is
        used for obtaining object properties as well as for the interval ratio
        correction of last_heads vectors."""
        self.scan += 1
        self.time = parse_grid_datetime(grid_obj1)
        if grid_obj2 is None:
            # tracks for last scan are being written
            return
        time2 = parse_grid_datetime(grid_obj2)
        old_diff = self.interval
        self.interval = cast(Union[timedelta, np.timedelta64], time2 - self.time)
        inter_val_s = get_interval(self.interval)
        if old_diff is not None:
            diff = get_interval(old_diff)
            self.interval_ratio = inter_val_s / diff


def get_interval(interval: Union[timedelta, np.timedelta64]) -> float:
    """Get a timeinterval in total seconds."""

    if isinstance(interval, np.timedelta64):
        return float(interval / np.timedelta64(1, "s"))
    return interval.total_seconds()


def convert_to_cftime(
    time: Union[np.datetime64, cftime.datetime]
) -> cftime.datetime:
    """Convert a dateim to a cftime object."""
    if isinstance(time, np.datetime64):
        t = datetime.fromisoformat(str(time).partition(".")[0])
        return cftime.DatetimeGregorian(
            t.year, t.month, t.day, t.hour, t.minute, t.second
        )
    return time


def get_grids(
    group: xr.Dataset,
    slices: tuple[int, int],
    lon: xr.DataArray,
    lat: xr.DataArray,
    times: xr.DataArray,
    varname: str = "rain_rate",
) -> Iterator[GridType]:

    dims = group[varname].dims
    for s in range(slices[0], slices[-1]):
        time = convert_to_cftime(
            cast(Union[np.datetime64, cftime.datetime], times[s].values)
        )
        mask_data = np.ma.masked_invalid(group[varname][s].values)
        if len(mask_data.shape) < 3:
            mask_data = mask_data[np.newaxis, :]
        yield GridType(
            x=group[dims[-1]],
            y=group[dims[-2]],
            lon=lon,
            lat=lat,
            data=mask_data,
            time=time,
        )



"""
tint.grid_utils
===============

Tools for pulling data from reading data.


"""

from __future__ import annotations
from typing import Union

import cftime
import numpy as np
import pandas as pd
from scipy import ndimage

from .types import ConfigType, GridType


def parse_grid_datetime(
    grid_obj: GridType,
) -> Union[np.datetime64, cftime.datetime]:
    """Obtains datetime object from the data dictionary."""
    return grid_obj.time


def get_grid_size(grid_obj: GridType) -> np.ndarray:
    """Calculates grid size per dimension given a grid object."""
    return np.array(grid_obj.data.shape)


def get_radar_info(radar: tuple[float, float]) -> dict[str, float]:
    """Return the longitude/latitude of the radar."""
    info = {"radar_lon": radar[0], "radar_lat": radar[1]}
    return info


def get_grid_alt(
    grid_size: tuple[float, float], alt_meters: Union[float, int] = 1500.0
) -> int:
    """Returns z-index closest to alt_meters."""
    return int(np.round(alt_meters / grid_size[0]))


def get_vert_projection(
    grid: np.ndarray, thresh: Union[float, int] = 40
) -> np.ndarray:
    """Returns boolean vertical projection from grid."""
    return np.any(grid > thresh, axis=0)


def get_filtered_frame(
    grid: np.ndarray, min_size: float, thresh: Union[int, float]
) -> np.ndarray:
    """Returns a labeled frame from gridded radar data. Smaller objects
    are removed and the rest are labeled."""
    echo_height = get_vert_projection(grid, thresh)
    labeled_echo = ndimage.label(echo_height)[0]
    frame = clear_small_echoes(labeled_echo, min_size)
    return frame


def clear_small_echoes(label_image: np.ndarray, min_size: float) -> np.ndarray:
    """Takes in binary image and clears objects less than min_size."""
    flat_image = pd.Series(label_image.flatten())
    flat_image = flat_image[flat_image > 0]
    size_table = flat_image.value_counts(sort=False)
    small_objects = size_table.keys()[size_table < min_size]

    for obj in small_objects:
        label_image[label_image == obj] = 0
    label_image = ndimage.label(label_image)
    return label_image[0]


def extract_grid_data(
    grid_obj: GridType, params: ConfigType
) -> tuple[np.ndarray, np.ndarray]:
    """Get filtered grid frame and raw grid slice at global shift altitude."""
    masked = grid_obj.data.filled(0)
    raw = masked[0, :, :]
    frame = get_filtered_frame(masked, params["MIN_SIZE"], params["FIELD_THRESH"])
    return raw, frame



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
TintX main tracking utility.

TintX offers a simple :class:``RunDirectory`` class that uses the ``xarray``
library to read datasets such as model output data or observational datasets.

The cell tracking as well as visualisation of cell tracks is realised through
instances of this :class:`RunDirectory` class.
"""

from tintx import config  # noqa: F401
from .reader import RunDirectory

__all__ = ["RunDirectory"]

__version__ = "2022.9.5"



"""Visualization tools for tracks objects."""

from __future__ import annotations
from functools import partial
from datetime import timedelta
from typing import Any, Optional, Iterator, Union

from cartopy import crs
from cartopy.mpl.geoaxes import GeoAxesSubplot
import pandas as pd
from matplotlib.animation import FuncAnimation
import matplotlib as mpl
from matplotlib import pyplot as plt
import numpy as np
from tqdm.auto import tqdm
import xarray as xr

from .tracks import Cell_tracks
from .types import GridType
from .helpers import convert_to_cftime


class Tracer:
    colors = ["m", "r", "lime", "darkorange", "k", "b", "darkgreen", "yellow"]
    colors.reverse()

    def __init__(self, tobj: Cell_tracks) -> None:
        self.tobj = tobj
        self.color_stack = self.colors * 10
        self.cell_color = pd.Series([], dtype=str)
        self.history: pd.DataFrame = pd.DataFrame()
        self.current: pd.DataFrame = pd.DataFrame()

    def update(self, nframe: int) -> None:
        self.history = self.tobj.tracks.loc[:nframe]
        self.current = self.tobj.tracks.loc[nframe]

    def _check_uid(self, uid: str) -> None:
        if uid not in self.cell_color.keys():
            try:
                self.cell_color[uid] = self.color_stack.pop()
            except IndexError:
                self.color_stack += self.colors * 5
                self.cell_color[uid] = self.color_stack.pop()

    def plot(self, ax: GeoAxesSubplot) -> None:
        for uid, group in self.history.groupby(level="uid"):
            self._check_uid(uid)
            tracer = group[["lon", "lat"]]
            if uid in self.current.index:
                ax.plot(tracer.lon, tracer.lat, self.cell_color[uid])


def _get_axes(
    X: Union[np.ndarray, xr.DataArray],
    Y: Union[np.ndarray, xr.DataArray],
    ax: Optional[GeoAxesSubplot],
    **kwargs: Union[float, int, str],
) -> GeoAxesSubplot:
    if ax is None:
        fig = plt.figure(figsize=(10, 8))
        ax = fig.add_subplot(111, projection=crs.PlateCarree())
    if not isinstance(ax, GeoAxesSubplot):
        raise TypeError("Ax should be a cartopy GeoAxesSubplot object")
    ax.set_ylim(Y.min(), Y.max())
    ax.set_xlim(X.min(), X.max())
    kwargs.setdefault("resolution", "10m")
    ax.coastlines(**kwargs)
    return ax


def _gen_from_grids(
    total_num: int, grid_0: GridType, grids: Iterator[GridType]
) -> Iterator[GridType]:

    with tqdm(total=total_num, desc="Animating", leave=False) as pbar:
        pbar.update()
        yield grid_0
        for grid in grids:
            pbar.update()
            yield grid


def full_domain(
    tobj: Cell_tracks,
    grids: Iterator[GridType],
    vmin: float = 0.01,
    vmax: float = 15,
    ax: Optional[GeoAxesSubplot] = None,
    cmap: Union[str, plt.cm] = "Blues",
    alt: Optional[float] = None,
    fps: float = 5,
    isolated_only: bool = False,
    tracers: bool = False,
    dt: float = 0,
    plot_style: Optional[dict[str, Union[float, int, str]]] = None,
) -> GeoAxesSubplot:
    alt = alt or tobj.params["GS_ALT"]
    plot_style = plot_style or {}
    shading = plot_style.pop("shading", "auto")
    if tracers:
        tracer = Tracer(tobj)
    nframes = tobj._tracks.index.levels[0].max()
    title = plot_style.pop("title", "")
    grid = next(grids)
    new_ax = _get_axes(grid.lon, grid.lat, ax, **plot_style)
    try:
        data = grid.data[0].filled(np.nan)
    except AttributeError:  # pragma: no cover
        data = grid.data[0]  # pragma: no cover
    im = new_ax.pcolormesh(
        grid.lon,
        grid.lat,
        data,
        vmin=vmin,
        vmax=vmax,
        cmap=cmap,
        shading=shading,
    )

    ann: dict[str, mpl.text.Annotation] = {}

    def _update(enum: tuple[int, GridType], title: str = "") -> None:
        for annotation in ann.values():
            try:
                annotation.remove()
            except ValueError:  # pragma: no cover
                pass  # pragma: no cover
        nframe, grid = enum
        try:
            im.set_array(grid.data[0].filled(np.nan).ravel())
        except AttributeError:  # pragma: no cover
            im.set_array(grid.data[0].ravel())  # pragma: no cover
        title_text = ""
        if title:
            title_text = f"{title} at "
        time_str = (convert_to_cftime(grid.time) + timedelta(hours=dt)).strftime(
            "%Y-%m-%d %H:%M"
        )
        new_ax.set_title(f"{title_text}{time_str}")
        if nframe in tobj._tracks.index.levels[0]:
            frame_tracks = tobj._tracks.loc[nframe]
            if tracers:
                tracer.update(nframe)
                tracer.plot(new_ax)
            for ind, uid in enumerate(frame_tracks.index):
                if isolated_only and not frame_tracks["isolated"].iloc[ind]:
                    continue
                x = frame_tracks["lon"].iloc[ind]
                y = frame_tracks["lat"].iloc[ind]
                ann[uid] = new_ax.annotate(uid, (x, y), fontsize=20)

    frames = enumerate(_gen_from_grids(nframes, grid, grids))
    animation = FuncAnimation(
        new_ax.get_figure(),
        partial(_update, title=title),
        frames=frames,
        interval=1000 / fps,
    )
    plt.close()
    return animation


def plot_traj(
    traj: pd.DataFrame,
    X: Union[np.ndarray, xr.DataArray],
    Y: Union[np.ndarray, xr.DataArray],
    label: bool = False,
    ax: Optional[GeoAxesSubplot] = None,
    uids: list[str] = None,
    mintrace: int = 2,
    size: int = 50,
    thresh: float = -1.0,
    color: Optional[str] = None,
    plot_style: Optional[dict[str, Union[float, int, str]]] = None,
) -> GeoAxesSubplot:

    """This code is a fork of plot_traj method in the plot module from the
    trackpy project see http://soft-matter.github.io/trackpy fro more details

    Plot traces of trajectories for each particle.

    Parameters
    ----------
    tobj : trajectory containing the tracking object
    X : 1D array of the X vector
    Y : 1D array of the Y vector
    label : boolean, default: False
        Set to True to write particle ID numbers next to trajectories.
    cmap : colormap, Default matplotlib.colormap.winter
        Colormap used to color different tracks
    ax: cartopy.mpl.geoaxes.GeoAxesSubplot, default: None
        Axes object that is used to create the animation. If None (default)
        a new axes object will be created.
    uids : list[str], default: None
    a preset of stroms to be drawn, instead of all
        (default)
    color : str, default: None
    A pre-defined color, if None (default) each track will be assigned a
    different color
    thresh : float, default: -1
        plot only trajectories with average intensity above this value.
    mintrace : int, default 2
        Minimum length of a trace to be plotted
    plot_style : dictionary
        Keyword arguments passed through to the `Axes.plot(...)` command

    Returns
    -------
    Axes object

    """

    plot_style = plot_style or {}
    _plot_style = dict(linewidth=1)
    resolution = plot_style.pop("resolution", "10m")
    _plot_style.update(**_normalize_kwargs(plot_style, "line2d"))
    size = _plot_style.pop("markersize", 50)
    ax = _get_axes(X, Y, ax, resolution=resolution, **_plot_style)
    proj = ax.projection
    if len(traj) == 0:
        raise ValueError("DataFrame of trajectories is empty.")

    # Read http://www.scipy.org/Cookbook/Matplotlib/MulticoloredLine
    y = traj["lat"]
    x = traj["lon"]
    val = traj["mean"]
    if uids is None:
        uid = np.unique(x.index.get_level_values("uid")).astype(np.int32)
        uid.sort()
        uids = uid.astype(str)

    for particle in uids:
        x1 = x[:, particle].values
        y1 = y[:, particle].values
        mean1 = val[:, particle].values.mean()
        if x1.shape[0] > int(mintrace) and mean1 >= thresh:
            im = ax.plot(x1, y1, color=color, **_plot_style)
            sc_color = im[0].get_color()
            ax.scatter(x1[0], y1[0], marker="o", color=sc_color, s=[size])
            ax.scatter(x1[-1], y1[-1], marker="*", color=sc_color, s=[size])
            if label:
                if len(x1) > 1:
                    cx, cy = proj.transform_point(
                        x1[int(x1.size / 2)], y1[int(y1.size / 2)], proj
                    )
                    dx, dy = proj.transform_point(
                        ((x1[1] - x1[0]) / 8.0), ((y1[1] - y1[0]) / 8.0), proj
                    )
                else:
                    cx, cy = proj.transform_point(x1[0], y1[0], proj)
                    dx, dy = 0, 0
                ax.annotate(
                    "%s" % str(particle),
                    xy=(cx, cy),
                    xytext=(cx + dx, cy + dy),
                    fontsize=16,
                    horizontalalignment="center",
                    verticalalignment="center",
                )
        else:
            im = None
    return ax


def _normalize_kwargs(
    kwargs: dict[str, Any], kind: str = "patch"
) -> dict[str, Any]:
    """Convert matplotlib keywords from short to long form."""
    # Source:
    # github.com/tritemio/FRETBursts/blob/fit_experim/fretbursts/burst_plot.py
    if kind == "line2d":
        long_names = dict(
            c="color",
            ls="linestyle",
            lw="linewidth",
            mec="markeredgecolor",
            mew="markeredgewidth",
            mfc="markerfacecolor",
            ms="markersize",
        )
    elif kind == "patch":
        long_names = dict(
            c="color",
            ls="linestyle",
            lw="linewidth",
            ec="edgecolor",
            fc="facecolor",
        )
    _ = kwargs.pop("resolution", None)
    for short_name in long_names.keys():
        if short_name in kwargs:
            kwargs[long_names[short_name]] = kwargs.pop(short_name)
    return kwargs



import os
from setuptools import setup, find_packages
import sys
from tempfile import TemporaryDirectory


def read(*parts):
    script_path = os.path.dirname(os.path.realpath(sys.argv[0]))
    return open(os.path.join(script_path, *parts)).read()


def find_version(*parts):

    vers_file = read(*parts).split("\n")
    old_path = sys.path.copy()
    with TemporaryDirectory() as td:
        with open(os.path.join(td, "tmp_frevaversion.py"), "w") as f:
            for line in vers_file:
                if "__version__" in line:
                    f.write(line)
        sys.path.insert(0, td)
        try:
            from tmp_frevaversion import __version__

            sys.path = old_path
            return __version__
        except ImportError:
            sys.path = old_path
    raise RuntimeError("Unable to find version string.")


meta = dict(
    description="Tracking facility to track rainfall and other non-continous data.",
    url="https://github.com/antarcticrainforest/tintX",
    author="Martin Bergemann",
    author_email="bergemann@dkrz.de",
    long_description=read("README.md"),
    include_package_data=True,
    long_description_content_type="text/markdown",
    license="BSD-3-Clause",
    python_requires=">=3.7",
    project_urls={
        "Documentation": "https://tintx.readthedocs.io/en/latest/",
        # 'Release notes': 'https://extra-data.readthedocs.io/en/latest/changelog.html',
        "Issues": "https://github.com/antarcticrainforest/tintX/issues",
        "Source": "https://github.com/antarcticrainforest/tintX",
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Atmospheric Science",
    ],
    version=find_version("src", "tintx", "__init__.py"),
    package_dir={"": "src"},
    entry_points={"console_scripts": ["tintx = tintx.cli:tintx"]},
    install_requires=[
        "cartopy",
        "click",
        "cftime",
        "dask",
        "pandas",
        "tables",
        "numpy",
        "scipy",
        "netCDF4",
        "matplotlib",
        "xarray",
        "tqdm",
        "typing_extensions",
    ],
    extras_require={
        "tests": [
            "bash_kernel",
            "black",
            "nbval",
            "pytest",
            "pytest-env",
            "pytest-cov",
            "testpath",
            "flake8",
            "mypy",
        ],
        "docs": [
            "bash_kernel",
            "black",
            "ipython",
            "ipywidgets",
            "h5netcdf",
            "mypy",
            "pytest",
            "recommonmark",
            "nbsphinx",
            "sphinx",
            "sphinxcontrib_github_alt",
            "sphinx-execute-code-python3",
            "sphinx-rtd-theme",
        ],
    },
)

setup(name="tintx", packages=find_packages("src"), **meta)



"""
tint.matching
=============

Functions for object matching between adjacent radar scans.

"""
from __future__ import annotations
from datetime import timedelta
from typing import cast, overload, Optional, Union

import numpy as np
import pandas as pd
from scipy import optimize

from .phase_correlation import get_ambient_flow
from .objects import get_obj_extent
from .helpers import get_interval, Record
from .types import ConfigType, SearchPredictType

LARGE_NUM = 1000


def euclidean_dist(vec1: np.ndarray, vec2: np.ndarray) -> float:
    """Computes euclidean distance."""
    vec1 = np.array(vec1)
    vec2 = np.array(vec2)
    dist = np.sqrt(sum((vec1 - vec2) ** 2))
    return dist


@overload
def get_size_change(size1: float, size2: float) -> float:
    ...  # pragma: no cover


@overload
def get_size_change(
    size1: Union[float, np.ndarray], size2: np.ndarray
) -> np.ndarray:
    ...  # pragma: no cover


@overload
def get_size_change(
    size1: np.ndarray, size2: Union[float, np.ndarray]
) -> np.ndarray:
    ...  # pragma: no cover


def get_size_change(
    size1: Union[np.ndarray, float], size2: Union[np.ndarray, float]
) -> Union[np.ndarray, float]:
    """Returns change in size of an echo as the ratio of the larger size to
    the smaller, minus 1."""
    if (size1 < 5) and (size2 < 5):
        return 0
    elif size1 >= size2:
        return size1 / size2 - 1
    else:
        return size2 / size1 - 1


def find_objects(search_box: SearchPredictType, image2: np.ndarray) -> np.ndarray:
    """Identifies objects found in the search region."""
    if not search_box["valid"]:
        obj_found = np.array(-1)
    else:
        search_area = image2[
            search_box["x1"] : search_box["x2"],
            search_box["y1"] : search_box["y2"],
        ]
        obj_found = np.unique(search_area)
    return obj_found


def shifts_disagree(
    shift1: float, shift2: float, record: Record, params: ConfigType
) -> bool:
    """Returns True if shift disparity greater than MAX_SHIFT_DISP
    parameter."""
    shift_1 = shift1 * record.grid_size[1:]
    shift_2 = shift2 * record.grid_size[1:]
    shift_disparity = euclidean_dist(shift_1, shift_2)
    interval = get_interval(
        cast(Union[timedelta, np.timedelta64], record.interval)
    )
    return shift_disparity / interval > params["MAX_SHIFT_DISP"]


def clip_shift(shift: float, record: Record, params: ConfigType) -> float:
    """Clips shift according to MAX_FLOW_MAG paramter."""
    shift_meters = shift * record.grid_size[1:]
    shift_mag = np.linalg.norm(shift_meters)
    interval = get_interval(
        cast(Union[timedelta, np.timedelta64], record.interval)
    )
    velocity = shift_mag / interval

    unit = shift_meters / shift_mag
    if velocity > params["MAX_FLOW_MAG"]:
        clipped = unit * params["MAX_FLOW_MAG"] * interval
        clipped_pix = clipped / record.grid_size[1:]
        return clipped_pix
    return shift


def correct_shift(
    local_shift: float,
    current_objects: Optional[pd.DataFrame],
    obj_id1: float,
    global_shift: float,
    record: Record,
    params: ConfigType,
) -> float:
    """Takes in flow vector based on local phase correlation (see
    get_std_flow) and compares it to the last headings of the object and
    the global_shift vector for that timestep. Corrects accordingly.
    Note: At the time of this function call, current_objects has not yet been
    updated for the current frame1 and frame2, so the id2s in current_objects
    correspond to the objects in the current frame1."""
    global_shift = clip_shift(global_shift, record, params)

    if current_objects is None:
        last_heads = None
    else:
        obj_index = current_objects["id2"] == obj_id1
        last_heads = current_objects["last_heads"][obj_index].flatten()
        last_heads = np.round(last_heads * record.interval_ratio, 2)
        if len(last_heads) == 0:
            last_heads = None

    if last_heads is None:
        if shifts_disagree(local_shift, global_shift, record, params):
            case = 0
            corrected_shift = global_shift
        else:
            case = 1
            corrected_shift = (local_shift + global_shift) / 2

    elif shifts_disagree(local_shift, last_heads, record, params):
        if shifts_disagree(local_shift, global_shift, record, params):
            case = 2
            corrected_shift = last_heads
        else:
            case = 3
            corrected_shift = local_shift

    else:
        case = 4
        corrected_shift = (local_shift + last_heads) / 2

    corrected_shift = np.round(corrected_shift, 2)

    record.count_case(case)
    record.record_shift(
        corrected_shift, global_shift, last_heads, local_shift, case
    )
    return corrected_shift


def predict_search_extent(
    obj1_extent: dict[str, np.ndarray],
    shift: float,
    params: ConfigType,
    grid_size: np.ndarray,
) -> SearchPredictType:
    """Predicts search extent/region for the object in image2 given
    the image shift."""
    shifted_center = obj1_extent["obj_center"] + shift
    search_radius_r = params["SEARCH_MARGIN"] / grid_size[1]
    search_radius_c = params["SEARCH_MARGIN"] / grid_size[2]
    x1 = shifted_center[0] - search_radius_r
    x2 = shifted_center[0] + search_radius_r + 1
    y1 = shifted_center[1] - search_radius_c
    y2 = shifted_center[1] + search_radius_c + 1
    x1 = int(x1)
    x2 = int(x2)
    y1 = int(y1)
    y2 = int(y2)
    return {
        "x1": x1,
        "x2": x2,
        "y1": y1,
        "y2": y2,
        "center_pred": shifted_center,
        "valid": True,
    }


def check_search_box(
    search_box: SearchPredictType, img_dims: tuple[int, ...]
) -> SearchPredictType:
    """Checks if search_box is within the boundaries of the frame. Clips to
    edges of frame if out of bounds. Marks as invalid if too small."""
    if search_box["x1"] < 0:
        search_box["x1"] = 0
    if search_box["y1"] < 0:
        search_box["y1"] = 0
    if search_box["x2"] > img_dims[0]:
        search_box["x2"] = img_dims[0]
    if search_box["y2"] > img_dims[1]:
        search_box["y2"] = img_dims[1]
    if (search_box["x2"] - search_box["x1"] < 5) or (
        search_box["y2"] - search_box["y1"] < 5
    ):
        search_box["valid"] = False
    return search_box


def get_disparity(
    obj_found: np.ndarray,
    image2: np.ndarray,
    search_box: SearchPredictType,
    obj1_extent: dict[str, np.ndarray],
) -> np.ndarray:
    """Computes disparities for objects in obj_found."""
    dist_pred = np.empty(0)
    change = np.empty(0)
    for target_obj in obj_found:
        target_extent = get_obj_extent(image2, target_obj)
        euc_dist = euclidean_dist(
            target_extent["obj_center"], search_box["center_pred"]
        )
        dist_pred = np.append(dist_pred, euc_dist)
        size_changed = get_size_change(
            target_extent["obj_area"], obj1_extent["obj_area"]
        )
        change = np.append(change, size_changed)

    disparity = dist_pred + change
    return disparity


def get_disparity_all(
    obj_found: np.ndarray,
    image2: np.ndarray,
    search_box: SearchPredictType,
    obj1_extent: dict[str, np.ndarray],
) -> np.ndarray:
    """Returns disparities of all objects found within the search box."""
    if np.max(obj_found) <= 0:
        disparity = np.array([LARGE_NUM])
    else:
        obj_found = obj_found[obj_found > 0]
        disparity = get_disparity(obj_found, image2, search_box, obj1_extent)
    return disparity


def save_obj_match(
    obj_id1: np.ndarray,
    obj_found: np.ndarray,
    disparity: np.ndarray,
    obj_match: np.ndarray,
    params: ConfigType,
) -> np.ndarray:
    """Saves disparity values in obj_match matrix. If disparity is greater than
    MAX_DISPARITY, saves a large number."""
    disparity[disparity > params["MAX_DISPARITY"]] = LARGE_NUM
    if np.max(obj_found) > 0:
        obj_found = obj_found[obj_found > 0]
        obj_found = obj_found - 1
        obj_id1 = obj_id1 - 1
        obj_match[obj_id1, obj_found] = disparity
    return obj_match


def locate_all_objects(
    image1: np.ndarray,
    image2: np.ndarray,
    global_shift: float,
    current_objects: Optional[pd.DataFrame],
    record: Record,
    params: ConfigType,
) -> Optional[np.ndarray]:
    """Matches all the objects in image1 to objects in image2. This is the
    main function called on a pair of images."""
    nobj1 = np.max(image1)
    nobj2 = np.max(image2)

    if (nobj2 == 0) or (nobj1 == 0):
        return None

    obj_match = np.full((nobj1, np.max((nobj1, nobj2))), LARGE_NUM, dtype="f")

    for obj_id1 in np.arange(nobj1) + 1:
        obj1_extent = get_obj_extent(image1, obj_id1)
        shift = get_ambient_flow(
            obj1_extent, image1, image2, params, record.grid_size
        )
        if shift is None:
            record.count_case(5)
            shift = global_shift

        shift = correct_shift(
            shift, current_objects, obj_id1, global_shift, record, params
        )

        search_box = predict_search_extent(
            obj1_extent, shift, params, record.grid_size
        )
        search_box = check_search_box(search_box, image2.shape)
        objs_found = find_objects(search_box, image2)
        disparity = get_disparity_all(objs_found, image2, search_box, obj1_extent)
        obj_match = save_obj_match(
            obj_id1, objs_found, disparity, obj_match, params
        )
    return obj_match


def match_pairs(obj_match: np.ndarray, params: ConfigType) -> np.ndarray:
    """Matches objects into pairs given a disparity matrix and removes
    bad matches. Bad matches have a disparity greater than the maximum
    threshold."""
    pairs = optimize.linear_sum_assignment(obj_match)

    for id1 in pairs[0]:
        if obj_match[id1, pairs[1][id1]] > params["MAX_DISPARITY"]:
            pairs[1][id1] = -1  # -1 indicates the object has died

    pairs = pairs[1] + 1  # ids in current_objects are 1-indexed
    return pairs


def get_pairs(
    image1: np.ndarray,
    image2: np.ndarray,
    global_shift: float,
    current_objects: Optional[dict[str, np.ndarray]],
    record: Record,
    params: ConfigType,
) -> Optional[np.ndarray]:
    """Given two images, this function identifies the matching objects and
    pairs them appropriately. See disparity function."""
    nobj1 = np.max(image1)
    nobj2 = np.max(image2)

    if nobj1 == 0:
        return None
    if nobj2 == 0:
        zero_pairs = np.zeros(nobj1)
        return zero_pairs

    obj_match = locate_all_objects(
        image1, image2, global_shift, current_objects, record, params
    )
    if obj_match is None:
        return None
    pairs = match_pairs(obj_match, params)
    return pairs



"""
tint.objects
============

Functions for managing and recording object properties.

"""
from __future__ import annotations
from typing import Union

import numpy as np
import pandas as pd
from scipy import ndimage
from .grid_utils import get_filtered_frame
from .helpers import Counter, Record
from .types import ConfigType, GridType, ObjectPropType


def get_object_center(
    obj_id: Union[int, float, str], labeled_image: np.ndarray
) -> int:
    """Returns index of center pixel of the given object id from labeled
    image. The center is calculated as the median pixel of the object extent;
    it is not a true centroid."""
    obj_index = np.argwhere(labeled_image == obj_id)
    center = np.median(obj_index, axis=0).astype("i")
    return center


def get_obj_extent(
    labeled_image: np.ndarray, obj_label: float
) -> dict[str, np.ndarray]:
    """Takes in labeled image and finds the radius, area, and center of the
    given object."""
    obj_index = np.argwhere(labeled_image == obj_label)

    xlength = np.max(obj_index[:, 0]) - np.min(obj_index[:, 0]) + 1
    ylength = np.max(obj_index[:, 1]) - np.min(obj_index[:, 1]) + 1
    obj_radius = np.max((xlength, ylength)) / 2
    obj_center = np.round(np.median(obj_index, axis=0), 0)
    obj_area = len(obj_index[:, 0])

    obj_extent = {
        "obj_center": obj_center,
        "obj_radius": obj_radius,
        "obj_area": obj_area,
        "obj_index": obj_index,
    }
    return obj_extent


def init_current_objects(
    first_frame: np.ndarray,
    second_frame: np.ndarray,
    pairs: np.ndarray,
    counter: Counter,
) -> tuple[dict[str, np.ndarray], Counter]:
    """Returns a dictionary for objects with unique ids and their
    corresponding ids in frame1 and frame1. This function is called when
    echoes are detected after a period of no echoes."""
    nobj = np.max(first_frame)

    id1 = np.arange(nobj) + 1
    uid = counter.next_uid(count=nobj)
    id2 = pairs
    obs_num = np.zeros(nobj, dtype="i")
    origin = np.array(["-1"] * nobj)

    current_objects = {
        "id1": id1,
        "uid": uid,
        "id2": id2,
        "obs_num": obs_num,
        "origin": origin,
    }
    current_objects = attach_last_heads(
        first_frame, second_frame, current_objects
    )
    return current_objects, counter


def update_current_objects(
    frame1: np.ndarray,
    frame2: np.ndarray,
    pairs: np.ndarray,
    old_objects: dict[str, np.ndarray],
    counter: Counter,
) -> tuple[dict[str, np.ndarray], Counter]:
    """Removes dead objects, updates living objects, and assigns new uids to
    new-born objects."""
    nobj = np.max(frame1)
    id1 = np.arange(nobj) + 1
    uid = np.array([], dtype="str")
    obs_num = np.array([], dtype="i")
    origin = np.array([], dtype="str")

    for obj in np.arange(nobj) + 1:
        if obj in old_objects["id2"]:
            obj_index = old_objects["id2"] == obj
            uid = np.append(uid, old_objects["uid"][obj_index])
            obs_num = np.append(obs_num, old_objects["obs_num"][obj_index] + 1)
            origin = np.append(origin, old_objects["origin"][obj_index])
        else:
            origin = np.append(origin, -1)
            uid = np.append(uid, counter.next_uid())
            obs_num = np.append(obs_num, 0)

    id2 = pairs
    current_objects = {
        "id1": id1,
        "uid": uid,
        "id2": id2,
        "obs_num": obs_num,
        "origin": origin,
    }
    return attach_last_heads(frame1, frame2, current_objects), counter


def attach_last_heads(
    frame1: np.ndarray, frame2: np.ndarray, current_objects: dict[str, np.ndarray]
) -> dict[str, np.ndarray]:
    """Attaches last heading information to current_objects dictionary."""
    nobj = len(current_objects["uid"])
    heads = np.ma.empty((nobj, 2))
    for obj in range(nobj):
        if (current_objects["id1"][obj] > 0) and (
            current_objects["id2"][obj] > 0
        ):
            center1 = get_object_center(current_objects["id1"][obj], frame1)
            center2 = get_object_center(current_objects["id2"][obj], frame2)
            heads[obj, :] = center2 - center1
        else:
            heads[obj, :] = np.ma.array([-999, -999], mask=[True, True])

    current_objects["last_heads"] = heads
    return current_objects


def check_isolation(
    raw: np.ndarray,
    filtered: np.ndarray,
    grid_size: np.ndarray,
    params: ConfigType,
) -> np.ndarray:
    """Returns list of booleans indicating object isolation. Isolated objects
    are not connected to any other objects by pixels greater than ISO_THRESH,
    and have at most one peak."""
    nobj = np.max(filtered)
    min_size = params["MIN_SIZE"] / np.prod(grid_size[1:] / 1000)
    iso_filtered = get_filtered_frame(raw, min_size, params["ISO_THRESH"])
    nobj_iso = np.max(iso_filtered)
    iso = np.empty(nobj, dtype="bool")

    for iso_id in np.arange(nobj_iso) + 1:
        obj_ind = np.where(iso_filtered == iso_id)
        objects = np.unique(filtered[obj_ind])
        objects = objects[objects != 0]
        if len(objects) == 1 and single_max(obj_ind, raw, params):
            iso[objects - 1] = True
        else:
            iso[objects - 1] = False
    return iso


def single_max(
    obj_ind: tuple[np.ndarray, ...], raw: np.ndarray, params: ConfigType
) -> bool:
    """Returns True if object has at most one peak."""
    max_proj = np.max(raw, axis=0)
    smooth = ndimage.gaussian_filter(max_proj, params["ISO_SMOOTH"])
    padded = np.pad(smooth, 1, mode="constant")
    obj_ind = tuple([axis + 1 for axis in obj_ind])  # adjust for padding
    maxima = 0
    for pixel in range(len(obj_ind[0])):
        ind_0 = obj_ind[0][pixel]
        ind_1 = obj_ind[1][pixel]
        neighborhood = padded[
            (ind_0 - 1) : (ind_0 + 2), (ind_1 - 1) : (ind_1 + 2)
        ]
        max_ind = np.unravel_index(neighborhood.argmax(), neighborhood.shape)
        if max_ind == (1, 1):
            maxima += 1
            if maxima > 1:
                return False
    return True


def get_object_prop(
    image1: np.ndarray,
    grid1: GridType,
    record: Record,
    params: ConfigType,
) -> ObjectPropType:
    """Returns dictionary of object properties for all objects found in
    image1."""
    id1: list[int] = []
    center: list[float] = []
    grid_x: list[float] = []
    grid_y: list[float] = []
    area: list[float] = []
    longitude: list[float] = []
    latitude: list[float] = []
    field_max: list[float] = []
    field_mean: list[float] = []
    nobj = np.max(image1)

    unit_area = 1  # (unit_dim[1]*unit_dim[2])/(1000**2)

    raw3D = grid1.data
    get_items = []
    for obj in np.arange(nobj) + 1:
        try:

            obj_index = np.argwhere(image1 == obj)
            this_centroid = np.round(np.mean(obj_index, axis=0), 3)
            rounded = np.round(this_centroid).astype("i")
            lon, lat = grid1.lon.values, grid1.lat.values
            if len(lon.shape) == 2:
                lon = lon[rounded[0], rounded[1]]
            else:
                lon = lon[rounded[1]]
            if len(lat.shape) == 2:
                lat = lat[rounded[0], rounded[1]]
            else:
                lat = lat[rounded[0]]
            longitude.append(np.round(lon, 4))
            latitude.append(np.round(lat, 4))

            id1.append(obj)

            # 2D frame stats
            center.append(np.median(obj_index, axis=0))
            grid_x.append(this_centroid[1])
            grid_y.append(this_centroid[0])
            area.append(obj_index.shape[0] * unit_area)

            # raw 3D grid stats
            obj_slices = [raw3D[:, ind[0], ind[1]] for ind in obj_index]
            field_max.append(float(np.nanmax(obj_slices)))
            field_mean.append(float(np.nanmean(obj_slices)))
            get_items.append(obj - 1)
        except IndexError:
            pass
    # cell isolation
    isolation = check_isolation(raw3D, image1, record.grid_size, params)
    objprop: ObjectPropType = {
        "id1": id1,
        "center": center,
        "grid_x": grid_x,
        "grid_y": grid_y,
        "area": area,
        "field_max": field_max,
        "field_mean": field_mean,
        "lon": longitude,
        "lat": latitude,
        "isolated": isolation,
        "ok_items": get_items,
    }
    return objprop


def write_tracks(
    old_tracks: pd.DataFrame,
    record: Record,
    current_objects: dict[str, np.ndarray],
    obj_props: ObjectPropType,
) -> pd.DataFrame:
    """Writes all cell information to tracks dataframe."""

    nobj = len(obj_props["id1"])
    scan_num = [record.scan] * nobj
    gi_obj = obj_props["ok_items"]
    uid = current_objects["uid"][gi_obj]
    new_tracks = pd.DataFrame(
        {
            "scan": scan_num,
            "uid": uid,
            "time": record.time,
            "grid_x": obj_props["grid_x"],
            "grid_y": obj_props["grid_y"],
            "lon": obj_props["lon"],
            "lat": obj_props["lat"],
            "area": obj_props["area"],
            "max": obj_props["field_max"],
            "mean": obj_props["field_mean"],
            "isolated": obj_props["isolated"][gi_obj],
        }
    )
    new_tracks.set_index(["scan", "uid"], inplace=True)
    tracks = pd.concat([old_tracks, new_tracks])
    return tracks



"""Collection of all type definitions used for type hinting."""
from __future__ import annotations
from typing import Union, NamedTuple, List
from typing_extensions import TypedDict

import cftime
import numpy as np
import xarray as xr

SearchPredictType = TypedDict(
    "SearchPredictType",
    {
        "x1": int,
        "x2": int,
        "y1": int,
        "y2": int,
        "center_pred": np.ndarray,
        "valid": bool,
    },
)
"""Type to store predictions for the objects in the next time step."""

ObjectPropType = TypedDict(
    "ObjectPropType",
    {
        "id1": List[int],
        "center": List[float],
        "grid_x": List[float],
        "grid_y": List[float],
        "area": List[float],
        "field_max": List[float],
        "field_mean": List[float],
        "lon": List[float],
        "lat": List[float],
        "isolated": np.ndarray,
        "ok_items": List[bool],
    },
)

CurrentObjectType = TypedDict(
    "CurrentObjectType",
    {
        "id1": np.ndarray,
        "uid": np.ndarray,
        "id2": np.ndarray,
        "obs_num": np.ndarray,
        "origin": np.ndarray,
    },
)
"""Parameter type describing properties of a tracked object."""

ConfigType = TypedDict(
    "ConfigType",
    {
        "FIELD_THRESH": float,
        "ISO_THRESH": float,
        "ISO_SMOOTH": float,
        "MIN_SIZE": float,
        "FLOW_MARGIN": float,
        "MAX_DISPARITY": float,
        "MAX_FLOW_MAG": float,
        "MAX_SHIFT_DISP": float,
        "SEARCH_MARGIN": float,
        "GS_ALT": float,
    },
)
"""The Parameter type defines the default values and types of the tint
tracking tuning parameters."""

GridType = NamedTuple(
    "GridType",
    [
        ("x", xr.DataArray),
        ("y", xr.DataArray),
        ("lon", xr.DataArray),
        ("lat", xr.DataArray),
        ("time", Union[np.datetime64, cftime.datetime]),
        ("data", np.ma.core.MaskedArray),
    ],
)
"""Type that describes the grid/time/data information for a time step."""



"""The :class:`RunDirectory` class is a convenience class to  access and apply
the tint tracking algoritm."""
from __future__ import annotations
from datetime import datetime
import hashlib
from pathlib import Path
from typing import Any, Optional, Union, Iterator
import warnings

from cartopy.mpl.geoaxes import GeoAxesSubplot
from matplotlib import pyplot as plt
import pandas as pd
from tqdm.auto import tqdm
import xarray as xr

from .helpers import get_grids, convert_to_cftime, MetaData
from .tracks import Cell_tracks
from .visualization import full_domain, plot_traj
from .config import set as set_config
from .types import GridType

__all__ = ["RunDirectory"]


class RunDirectory(Cell_tracks):
    """
    Create an :class:`RunDirectory` object from a given xarray dataset.

    The :class:`RunDirectory` object gathers all necessary information of
    data that is stored in a ``xarray`` dataset. Once loaded the most
    important meta data will be stored in the run directory for faster
    access the second time.

    Parameters
    ----------
    dataset: xarray.DataArray, xarray.Dataset
        Input dataset/data array
    var_name: str
        Name of the variable that is tracked
    x_coord: str, default: lon
        Name of the X coordinate array/vector
    y_coord: str, default: lat
        Name of the Y coordinate array/vector
    time_coord: str, default: time
        Name of the time variable.

    Example
    -------
    .. execute_code::
        :hide_headers:

        import os
        import xarray
        from tintx import RunDirectory
        files = os.path.join(os.environ["FILE_PATH"], "CMORPH*.nc")
        dset = xarray.open_mfdataset(files, combine="by_coords")
        run = RunDirectory(dset.isel(time=slice(10, 150)), "prcip")

    """

    _files: Union[str, list[str]] = ""
    _parameters: dict[str, dict[str, float]] = {}
    """Nested dict of tracking parameters that are saved to file."""

    @classmethod
    def from_files(
        cls,
        input_files: Union[Path, str, list[str], list[Path], Iterator[Path]],
        var_name: str,
        *,
        x_coord: str = "lon",
        y_coord: str = "lat",
        time_coord: str = "time",
        start: Optional[Union[str, datetime, pd.Timestamp]] = None,
        end: Optional[Union[str, datetime, pd.Timestamp]] = None,
        **kwargs: Any,
    ) -> RunDirectory:
        """
        Create a :class:`RunDirectory` object from input file(s)/directory.

        The :class:`RunDirectory` object gathers all necessary information of
        the data that is stored in the run directory. Once loaded the most
        important meta data will be stored in the run directory for faster
        access the second time.

        Parameters
        ----------
        inp_files: os.PathLike, list[os.PathLike]
            Input filenames or directory that is opened.
        var_name: str
            Name of the variable that is tracked
        start: str, pandas.Timestamp, datetime.datetime (default: None)
            first time step that is considered, if None given the first
            time step in the data is considered.
        end: str, pandas.Timestamp, datetime.datetime (default: None)
            last time step that is considered, if None given the last
            time step in the data is considered.
        x_coord: str (default: lon)
            The name of the longitude vector/array, can be 1D or 2D
        x_coord: str (default: lat)
            The name of the latitude vector/array, can be 1D or 2D
        time_coord: str, default: time
            The name of the time variable
        kwargs:
            Additional keyword arguments that are passed to open the dataset
            with xarray

        Example
        -------
        .. execute_code::
            :hide_headers:

            import os
            from tintx import RunDirectory
            run = RunDirectory.from_files(
                os.path.join(os.environ["FILE_PATH"], "CMORPH*.nc"),
                "cmorph"
            )

        """
        defaults: dict[str, Union[str, bool]] = dict(
            combine="by_coords",
        )
        _files: Union[str, list[str]] = ""
        if isinstance(input_files, (str, Path)):
            _files = str(input_files)
        else:
            _files = [str(f) for f in input_files]
        for key, value in defaults.items():
            kwargs.setdefault(key, value)
        _dset = xr.open_mfdataset(_files, **kwargs)
        start_time = start or _dset[time_coord].isel({time_coord: 0})
        end_time = end or _dset[time_coord].isel({time_coord: -1})
        return cls(
            _dset.sel({time_coord: slice(start_time, end_time)}),
            var_name,
            x_coord=x_coord,
            y_coord=y_coord,
            time_coord=time_coord,
            _files=_files,
        )

    def __init__(
        self,
        dataset: Union[xr.Dataset, xr.DataArray],
        var_name: str,
        *,
        time_coord: str = "time",
        x_coord: str = "lon",
        y_coord: str = "lat",
        _files: Union[list[str], str] = "",
    ) -> None:
        if isinstance(dataset, xr.DataArray):
            self.data = xr.Dataset({var_name: dataset})
        else:
            self.data = dataset
        self._files = _files
        self.lons = self.data[x_coord]
        self.lats = self.data[y_coord]
        self.var_name = var_name
        self.time = self.data[time_coord]
        self.start = convert_to_cftime(self.time.values[0])
        self.end = convert_to_cftime(self.time.values[-1])
        self._metadata_reader = MetaData(
            self.data,
            self.var_name,
            {"x_coord": x_coord, "y_coord": y_coord, "time_coord": time_coord},
        )
        super().__init__(var_name)

    def get_tracks(
        self,
        centre: Optional[tuple[float, float]] = None,
        leave_bar: bool = True,
        flush: bool = True,
        **tracking_parameters: float,
    ) -> int:
        """Apply the ``tint`` tracking algorithm.

        This is the primary method of the :class:`RunDirectory` class. This
        methods applies the tracking algorithm to the data and saves
        the tracked cells to a pandas ``DataFrame`` for analysis.

        Parameters
        ----------
        centre: tuple, default: None
            The centre of the radar station
        leave_bar: bool, default: True
            Leave the progress bar after tracking is finished
        flush: bool, default: True
            Flush old tracking data. If false, the tracks will be added
            to the existing ::method::`tracks` DataFrame.
        **tracking_parameters: float
            Overwrite the tint tracking parameters with this values for this
            specific tracking. The defaults will be restored afterwards.
            See :py:mod:`tintx.config` for details


        See also
        --------
        :py:mod:`tintx.config`

        Returns
        -------
        int: Number of unique cells identified by the tracking

        Example
        -------
        .. execute_code::
            :hide_headers:

            import os
            import xarray
            from tintx import RunDirectory
            files = os.path.join(os.environ["FILE_PATH"], "CPOL*.nc")
            dset = xarray.open_mfdataset(files, combine="by_coords")
            run = RunDirectory(dset,
                               "radar_estimated_rain_rate",
                               x_coord="longitude",
                               y_coord="latitude")
            num_cells = run.get_tracks(field_thresh=0.01)

        """
        if flush:
            self.reset_tracks()
        parameters = self._parameters.get(self._track_hash(), {})
        parameters.update(tracking_parameters)
        with tqdm(
            self.grids, total=self.time.size - 1, desc="Tracking", leave=leave_bar
        ) as pbar:
            with set_config(**parameters) as cfg:
                num_tracks = self._get_tracks(self.grids, pbar, centre)
                self._parameters[self._track_hash()] = cfg.config.copy()
            return num_tracks

    @property
    def grids(self) -> Iterator[GridType]:
        """Iterator holding longitude/latitude/time and data."""
        yield from get_grids(
            self.data,
            (0, self.time.size),
            self.lons,
            self.lats,
            self.time,
            varname=self.var_name,
        )

    @property
    def tracks(self) -> pd.DataFrame:
        """Pandas ``DataFrame`` representation of the tracked cells."""
        return self._tracks

    def save_tracks(self, output: Union[str, Path]) -> None:
        """Save tracked data to hdf5 table.

        Parameters
        ----------
        output: Path, str
            Filename where the output will be saved to.

        Example
        -------
        .. execute_code::
            :hide_headers:

            import os
            import xarray
            from tintx import RunDirectory
            files = os.path.join(os.environ["FILE_PATH"], "CPOL*.nc")
            dset = xarray.open_mfdataset(files, combine="by_coords")
            run = RunDirectory(dset,
                               "radar_estimated_rain_rate",
                               x_coord="longitude",
                               y_coord="latitude")
            num_cells = run.get_tracks(field_thresh=0.01)
            run.save_tracks("/tmp/output.hdf5")

        See Also
        --------
        :class:`from_dataframe`
        """
        output = Path(output).expanduser().absolute()
        metadata = self._metadata.copy()
        metadata["files"] = self._files
        with pd.HDFStore(output, "w") as hdf5:
            with warnings.catch_warnings():
                warnings.simplefilter(
                    action="ignore", category=pd.errors.PerformanceWarning
                )
                hdf5.put("tintx_tracks", self.tracks)
                self._metadata_reader.save(hdf5)
            table = hdf5.get_storer("tintx_tracks")
            table.attrs.track_meta = metadata
            table.attrs.tracking_parameters = self._parameters[self._track_hash()]

    @classmethod
    def from_dataframe(
        cls, track_file: Union[str, Path], dataset: Optional[xr.Dataset] = None
    ) -> RunDirectory:
        """Create an instance of the :class:`RunDirectory` class from tintx tracks.

        Parameters
        ----------
        track_file: str, pathlib.Path
            Path to the hdf5 file containing the tint tracks
        dataset: xr.Dataset, default: None
            xarray Dataset holding the data of the tracked field.
            if None (default) the dataset will be re opened from the files.
            Note: this only works if the previous instance of the tracking
            class was instanciated with the :class:`from_files` method.

        .. note::
        If the dataset can't be opened an empty data set will be created
        instead.

        Example
        -------
        .. execute_code::
            :hide_code:
            :hide_headers:

            from pathlib import Path
            if not Path("/tmp/output.hdf5").is_file():
                from tintx import RunDirectory
                run = RunDirectory.from_files(
                    os.path.join(os.environ["FILE_PATH"], "CPOL*.nc"),
                    "radar_estimated_rain_rate", x_coord="x", y_coord="y"
                )
                run.get_tracks(min_size=4, field_thresh=0.1)
                run.save_tracks("/tmp/output.hdf5")
                run = RunDirectory.from_dataframe("/tmp/output.hdf5")

        ::

            from tintx import RunDirectory
            run = RunDirectory.from_dataframe("/tmp/output.hdf5")

        See Also
        --------
        :class:`save_tracks`
        """

        track_file = Path(track_file).expanduser().absolute()
        with pd.HDFStore(track_file) as hdf5:
            tracks = pd.read_hdf(hdf5, "tintx_tracks")
            metadata = hdf5.get_storer("tintx_tracks").attrs.track_meta.copy()
            parameters = hdf5.get_storer("tintx_tracks").attrs.tracking_parameters
            coord_dataset = MetaData.dataset_from_coords(hdf5)

        files = metadata.pop("files", "")
        var_name = metadata.pop("var_name", "")
        start = metadata.pop("start", None) or None
        end = metadata.pop("end", None) or None
        try:
            if dataset is None:
                cls_instance = cls.from_files(
                    files,
                    var_name,
                    start=start,
                    end=end,
                    **metadata,
                )
            else:
                cls_instance = cls(dataset, var_name, _files=files, **metadata)
        except Exception as error:
            warnings.warn(
                "Could not access original data, creating empty dataset "
                f"the reason for the failure was:\n{error}"
            )
            cls_instance = cls(coord_dataset, var_name, **metadata)
        cls_instance._metadata[cls_instance._track_hash()] = parameters
        cls_instance.reset_tracks(tracks)
        return cls_instance

    def get_parameters(
        self, tracks: Optional[pd.DataFrame] = None
    ) -> dict[str, float]:
        """Get the parameters of given cell tracks.

        Parameters
        ----------
        tracks: pd.DataFrame, default: None
            The tracks ``DataFrame`` that is the result of the tuning
            parameters in question.

        Returns
        -------
        dict: dictionary holding the tuning parameters for the given tracks.

        Raises
        ------
        ValueError: if no parameters matching the input tracks ``DataFrame``
                    could be found.

        Example
        -------
        .. execute_code::
            :hide_code:
            :hide_headers:

            from pathlib import Path
            if not Path("/tmp/output.hdf5").is_file():
                from tintx import RunDirectory
                run = RunDirectory.from_files(
                    os.path.join(os.environ["FILE_PATH"], "CPOL*.nc"),
                    "radar_estimated_rain_rate", x_coord="x", y_coord="y"
                )
                run.get_tracks(min_size=4, field_thresh=0.1)
                run.save_tracks("/tmp/output.hdf5")
                parameters = run.get_parameters()

        ::

            from tintx import RunDirectory
            run = RunDirectory.from_dataframe("/tmp/output.hdf5")
            parameters = run.get_parameters()
        """

        try:
            return self._parameters[self._track_hash(tracks)]
        except KeyError as error:
            raise ValueError(
                "Could not retrieve parameters for given tracks"
            ) from error

    def _track_hash(self, tracks: Optional[pd.DataFrame] = None) -> str:
        if tracks is None:
            tracks = self._tracks
        df_hash = hashlib.sha512(tracks.to_string().encode("utf-8"))
        return df_hash.hexdigest()

    @property
    def _metadata(self) -> dict[str, Any]:
        return dict(
            x_coord=self.lons.name,
            y_coord=self.lats.name,
            time_coord=self.time.name,
            var_name=self.var_name,
            start=str(self.start),
            end=str(self.end),
        )

    def reset_tracks(self, tracks: Optional[pd.DataFrame] = None) -> None:
        """Override the tack data with a given DataFrame.

        Parameters
        ----------
        tracks: pd.DataFrame, default: None
            Tracking information used to override the current tarcks.
            If None (default) an empty DataFrame will be used.

        Example
        -------
        .. execute_code::
            :hide_headers:

            import os
            import xarray
            from tintx import RunDirectory
            files = os.path.join(os.environ["FILE_PATH"], "CPOL*.nc")
            dset = xarray.open_mfdataset(files, combine="by_coords")
            run = RunDirectory(dset,
                               "radar_estimated_rain_rate",
                               x_coord="longitude",
                               y_coord="latitude")
            num_cells = run.get_tracks(field_thresh=0.01)
            print(len(run.tracks))
            run.reset_tracks()
            print(len(run.tracks))
        """
        self.record = None
        self.counter = None
        self.current_objects = None
        if tracks is None:
            self._tracks = pd.DataFrame()
        else:
            self._tracks = tracks
        self._save()

    def animate(
        self,
        vmin: float = 0.01,
        vmax: float = 15,
        ax: Optional[GeoAxesSubplot] = None,
        cmap: Union[str, plt.cm] = "Blues",
        alt: Optional[float] = None,
        fps: float = 5,
        isolated_only: bool = False,
        tracers: bool = False,
        dt: float = 0,
        plot_style: Optional[dict[str, Union[float, int, str]]] = None,
    ) -> GeoAxesSubplot:
        """
        Create a animation of tracked cells.

        Parameters
        ----------
        vmin : float, default: 0.01
            Minimum values for the colormap.
        vmax : float, default: 15.0
            Maximum values for the colormap.
        isolated_only: bool, default: False
            If true, only annotates uids for isolated objects.
        cmap: str, default: Blues
            Colormap used for plotting the tracked fields.
        alt: float, default: None
            Altitude to be plotted in meters, for 3D data only.
        tracers: bool, default: False
            Plot traces of animated cells
        dt: float, default: 0
            Time shift in hours that is applied to the data, this can be
            useful if time data is in utc but should be displayed in
            another time zone.
        fps: int, default: 5
            Frames per second for output.
        ax: cartopy.mpl.geoaxes.GeoAxesSubplot, default: None
            Axes object that is used to create the animation. If None (default)
            a new axes object will be created.
        plot_style: dict
            Additional keyword arguments passed to the plotting routine.

        Returns
        -------
        matplotlib.animation.FuncAnimation:
            See `FuncAnimation <https://matplotlib.org/stable/api/_as_gen/matplotlib.animation.FuncAnimation.html>`_
            for more details.

        Example
        -------

        .. execute_code::
            :hide_headers:

            from tintx import RunDirectory
            run = RunDirectory.from_files(
                os.path.join(os.environ["FILE_PATH"], "CPOL*.nc"),
                "radar_estimated_rain_rate", x_coord="x", y_coord="y"
            )
            run.get_tracks(min_size=4, field_thresh=2)
            anim = run.animate(vmax=3, fps=2, plot_style={"res": "10m", "lw":1})

        """
        return full_domain(
            self,
            self.grids,
            vmin=vmin,
            vmax=vmax,
            ax=ax,
            cmap=cmap,
            alt=alt,
            fps=fps,
            isolated_only=isolated_only,
            tracers=tracers,
            dt=dt,
            plot_style=plot_style,
        )

    def plot_trajectories(
        self,
        label: bool = False,
        ax: Optional[GeoAxesSubplot] = None,
        uids: list[str] = None,
        mintrace: int = 2,
        size: int = 50,
        thresh: float = -1.0,
        color: Optional[str] = None,
        plot_style: Optional[dict[str, Union[float, int, str]]] = None,
    ) -> GeoAxesSubplot:
        """
        Plot traces of trajectories for each cell track.

        This code is a fork of ``plot_traj`` method in the plot module from the
        trackpy project see http://soft-matter.github.io/trackpy for more
        details

        Parameters
        ----------
        label : boolean, default: False
            Set to True to write cell uids next to trajectories.
        cmap : colormap, Default matplotlib.colormap.winter
            Colormap used to color different tracks
        ax: cartopy.mpl.geoaxes.GeoAxesSubplot, default: None
            Axes object that is used to create the animation. If None (default)
            a new axes object will be created.
        uids : list[str], default: None
            a preset of stroms to be drawn, instead of all (default)
        color : str, default: None
            A pre-defined color, if None (default) each track will be assigned
            a different color.
        thresh : float, default: -1
            Plot only trajectories with average intensity above this value.
        mintrace : int, default 2
            Minimum length of a trace to be plotted
        plot_style: dict
            Additional keyword arguments passed through to the
            ``Axes.plot(...)``

        Returns
        -------
        cartopy.mpl.geoaxes.GeoAxesSubplot


        Example
        -------
        .. execute_code::
            :hide_code:
            :hide_headers:

            from pathlib import Path
            if not Path("/tmp/output.hdf5").is_file():
                from tintx import RunDirectory
                run = RunDirectory.from_files(
                    os.path.join(os.environ["FILE_PATH"], "CPOL*.nc"),
                    "radar_estimated_rain_rate", x_coord="x", y_coord="y"
                )
                run.get_tracks(min_size=4, field_thresh=0.1)
                run.save_tracks("/tmp/output.hdf5")
                run = RunDirectory.from_dataframe("/tmp/output.hdf5")

        ::

            from tintx import RunDirectory
            run = RunDirectory.from_dataframe("/tmp/output.hdf5")
            ax = run.plot_trajectories(thresh=2, plot_style={"ms":25, "lw":1})
        """
        return plot_traj(
            self.tracks,
            self.lons,
            self.lats,
            ax=ax,
            label=label,
            uids=uids,
            mintrace=mintrace,
            size=size,
            thresh=thresh,
            color=color,
            plot_style=plot_style,
        )
